"""Testing copyfile functions."""
# Authors: Chris Holdgraf <choldgraf@berkeley.edu>
#          Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#
# License: BSD-3-Clause
import os.path as op
import datetime
from pathlib import Path

import pytest

import mne
from mne.datasets import testing
from mne_bids import BIDSPath
from mne_bids.path import _parse_ext
from mne_bids.copyfiles import (_get_brainvision_encoding,
                                _get_brainvision_paths,
                                copyfile_brainvision,
                                copyfile_edf,
                                copyfile_eeglab,
                                copyfile_kit)

testing_path = testing.data_path(download=False)
base_path = op.join(op.dirname(mne.__file__), 'io')


@testing.requires_testing_data
def test_get_brainvision_encoding():
    """Test getting the file-encoding from a BrainVision header."""
    data_path = op.join(base_path, 'brainvision', 'tests', 'data')
    raw_fname = op.join(data_path, 'test.vhdr')

    with pytest.raises(UnicodeDecodeError):
        with open(raw_fname, 'r', encoding='ascii') as f:
            f.readlines()

    enc = _get_brainvision_encoding(raw_fname)
    with open(raw_fname, 'r', encoding=enc) as f:
        f.readlines()


def test_get_brainvision_paths(tmp_path):
    """Test getting the file links from a BrainVision header."""
    data_path = op.join(base_path, 'brainvision', 'tests', 'data')
    raw_fname = op.join(data_path, 'test.vhdr')

    with pytest.raises(ValueError):
        _get_brainvision_paths(op.join(data_path, 'test.eeg'))

    # Write some temporary test files
    with open(tmp_path / 'test1.vhdr', 'w') as f:
        f.write('DataFile=testing.eeg')

    with open(tmp_path / 'test2.vhdr', 'w') as f:
        f.write('MarkerFile=testing.vmrk')

    with pytest.raises(ValueError):
        _get_brainvision_paths(tmp_path / 'test1.vhdr')

    with pytest.raises(ValueError):
        _get_brainvision_paths(tmp_path / 'test2.vhdr')

    # This should work
    eeg_file_path, vmrk_file_path = _get_brainvision_paths(raw_fname)
    head, tail = op.split(eeg_file_path)
    assert tail == 'test.eeg'
    head, tail = op.split(vmrk_file_path)
    assert tail == 'test.vmrk'


@pytest.mark.filterwarnings('ignore:.*Exception ignored.*:'
                            'pytest.PytestUnraisableExceptionWarning')
def test_copyfile_brainvision(tmp_path):
    """Test the copying of BrainVision vhdr, vmrk and eeg files."""
    bids_root = str(tmp_path)
    data_path = op.join(base_path, 'brainvision', 'tests', 'data')
    raw_fname = op.join(data_path, 'test.vhdr')
    new_name = op.join(bids_root, 'tested_conversion.vhdr')

    # IO error testing
    with pytest.raises(ValueError, match='Need to move data with same'):
        copyfile_brainvision(raw_fname, new_name + '.eeg')

    # Try to copy the file
    copyfile_brainvision(raw_fname, new_name)

    # Have all been copied?
    head, tail = op.split(new_name)
    assert op.exists(op.join(head, 'tested_conversion.vhdr'))
    assert op.exists(op.join(head, 'tested_conversion.vmrk'))
    assert op.exists(op.join(head, 'tested_conversion.eeg'))

    # Try to read with MNE - if this works, the links are correct
    raw = mne.io.read_raw_brainvision(new_name)
    assert raw.filenames[0] == (op.join(head, 'tested_conversion.eeg'))

    # Test with anonymization
    raw = mne.io.read_raw_brainvision(raw_fname)
    prev_date = raw.info['meas_date']
    anonymize = {'daysback': 32459}
    copyfile_brainvision(raw_fname, new_name, anonymize)
    raw = mne.io.read_raw_brainvision(new_name)
    new_date = raw.info['meas_date']
    assert new_date == (prev_date - datetime.timedelta(days=32459))


def test_copyfile_edf(tmp_path):
    """Test the anonymization of EDF/BDF files."""
    bids_root = tmp_path / "bids1"
    bids_root.mkdir()
    data_path = op.join(base_path, 'edf', 'tests', 'data')

    # Test regular copying
    for ext in ['.edf', '.bdf']:
        raw_fname = op.join(data_path, 'test' + ext)
        new_name = op.join(bids_root, 'test_copy' + ext)
        copyfile_edf(raw_fname, new_name)

    # IO error testing
    with pytest.raises(ValueError, match='Need to move data with same'):
        raw_fname = op.join(data_path, 'test.edf')
        new_name = op.join(bids_root, 'test_copy.bdf')
        copyfile_edf(raw_fname, new_name)

    # Add some subject info to an EDF to test anonymization
    testfile = op.join(bids_root, 'test_copy.edf')
    raw_date = mne.io.read_raw_edf(testfile).info['meas_date']
    date = datetime.datetime.strftime(raw_date, "%d-%b-%Y").upper()
    test_id_info = '023 F 02-AUG-1951 Jane'
    test_rec_info = 'Startdate {0} ID-123 John BioSemi_ActiveTwo'.format(date)
    with open(testfile, 'r+b') as f:
        f.seek(8)
        f.write(bytes(test_id_info.ljust(80), 'ascii'))
        f.write(bytes(test_rec_info.ljust(80), 'ascii'))

    # Test date anonymization
    def _edf_get_real_date(fpath):
        with open(fpath, 'rb') as f:
            f.seek(88)
            rec_info = f.read(80).decode('ascii').rstrip()
        startdate = rec_info.split(' ')[1]
        return datetime.datetime.strptime(startdate, "%d-%b-%Y")

    bids_root2 = tmp_path / 'bids2'
    bids_root2.mkdir()
    infile = op.join(bids_root, 'test_copy.edf')
    outfile = op.join(bids_root2, 'test_copy_anon.edf')
    anonymize = {'daysback': 33459, 'keep_his': False}
    copyfile_edf(infile, outfile, anonymize)
    new_date = _edf_get_real_date(outfile)

    # new anonymized date should be the minimum in EDF spec
    # (i.e. 01-01-1985)
    assert new_date == datetime.datetime(year=1985, month=1, day=1)

    # Test full ID info anonymization
    anon_startdate = datetime.datetime.strftime(new_date, "%d-%b-%Y").upper()
    with open(outfile, 'rb') as f:
        f.seek(8)
        id_info = f.read(80).decode('ascii').rstrip()
        rec_info = f.read(80).decode('ascii').rstrip()
    rec_info_tmp = "Startdate {0} X mne-bids_anonymize X"
    assert id_info == "0 X X X"
    assert rec_info == rec_info_tmp.format(anon_startdate)

    # Test partial ID info anonymization
    outfile2 = op.join(bids_root2, 'test_copy_anon_partial.edf')
    anonymize = {'daysback': 33459, 'keep_his': True}
    copyfile_edf(infile, outfile2, anonymize)
    with open(outfile2, 'rb') as f:
        f.seek(8)
        id_info = f.read(80).decode('ascii').rstrip()
        rec_info = f.read(80).decode('ascii').rstrip()
    rec = 'Startdate {0} ID-123 John BioSemi_ActiveTwo'.format(anon_startdate)
    assert id_info == "023 F X X"
    assert rec_info == rec


def test_copyfile_edfbdf_uppercase(tmp_path):
    """Test the copying of EDF/BDF files with upper-case extension."""
    bids_root = tmp_path / "bids1"
    bids_root.mkdir()
    data_path = op.join(base_path, 'edf', 'tests', 'data')

    # Test regular copying
    for ext in ['.edf', '.bdf']:
        raw_fname = op.join(data_path, 'test' + ext)
        new_name = op.join(bids_root, 'test_copy' + ext.upper())

        with pytest.warns(RuntimeWarning, match='Upper-case extension'):
            copyfile_edf(raw_fname, new_name)
        assert Path(new_name).with_suffix(ext).exists()


@pytest.mark.parametrize('fname',
                         ('test_raw.set', 'test_raw_chanloc.set',
                          'test_raw_2021.set'))
@testing.requires_testing_data
def test_copyfile_eeglab(tmp_path, fname):
    """Test the copying of EEGlab set and fdt files."""
    bids_root = str(tmp_path)
    data_path = op.join(testing_path, 'EEGLAB')
    raw_fname = op.join(data_path, fname)
    new_name = op.join(bids_root, f'CONVERTED_{fname}.set')

    # IO error testing
    with pytest.raises(ValueError, match="Need to move data with same ext"):
        copyfile_eeglab(raw_fname, new_name + '.wrong')

    # Test copying and reading
    copyfile_eeglab(raw_fname, new_name)
    if fname == 'test_raw_chanloc.set':  # combined set+fdt
        with pytest.warns(RuntimeWarning,
                          match="The data contains 'boundary' events"):
            raw = mne.io.read_raw_eeglab(new_name)
            assert 'Fp1' in raw.ch_names
    else:  # combined set+fdt and single set (new EEGLAB format)
        raw = mne.io.read_raw_eeglab(new_name, preload=True)
        assert 'EEG 001' in raw.ch_names
    assert isinstance(raw, mne.io.BaseRaw)


def test_copyfile_kit(tmp_path):
    """Test copying and renaming KIT files to a new location."""
    output_path = str(tmp_path)
    data_path = op.join(base_path, 'kit', 'tests', 'data')
    raw_fname = op.join(data_path, 'test.sqd')
    hpi_fname = op.join(data_path, 'test_mrk.sqd')
    electrode_fname = op.join(data_path, 'test.elp')
    headshape_fname = op.join(data_path, 'test.hsp')
    subject_id = '01'
    session_id = '01'
    run = '01'
    acq = '01'
    task = 'testing'

    raw = mne.io.read_raw_kit(
        raw_fname, mrk=hpi_fname, elp=electrode_fname,
        hsp=headshape_fname)
    _, ext = _parse_ext(raw_fname)

    datatype = 'meg'  # copyfile_kit makes the same assumption

    bids_path = BIDSPath(
        subject=subject_id, session=session_id, run=run, acquisition=acq,
        task=task)
    kit_bids_path = bids_path.copy().update(acquisition=None,
                                            datatype=datatype,
                                            root=output_path)
    bids_fname = str(bids_path.copy().update(datatype=datatype,
                                             suffix=datatype,
                                             extension=ext,
                                             root=output_path))

    copyfile_kit(raw_fname, bids_fname, subject_id, session_id,
                 task, run, raw._init_kwargs)
    assert op.exists(bids_fname)
    _, ext = _parse_ext(hpi_fname)
    if ext == '.sqd':
        kit_bids_path.update(suffix='markers', extension='.sqd')
        assert op.exists(kit_bids_path)
    elif ext == '.mrk':
        kit_bids_path.update(suffix='markers', extension='.mrk')
        assert op.exists(kit_bids_path)

    if op.exists(electrode_fname):
        task, run, key = None, None, 'ELP'
        elp_ext = '.pos'
        elp_fname = BIDSPath(
            subject=subject_id, session=session_id, task=task, run=run,
            acquisition=key, suffix='headshape', extension=elp_ext,
            datatype='meg', root=output_path)
        assert op.exists(elp_fname)

    if op.exists(headshape_fname):
        task, run, key = None, None, 'HSP'
        hsp_ext = '.pos'
        hsp_fname = BIDSPath(
            subject=subject_id, session=session_id, task=task, run=run,
            acquisition=key, suffix='headshape', extension=hsp_ext,
            datatype='meg', root=output_path)
        assert op.exists(hsp_fname)



"""Write Elekta/Neuromag/MEGIN crosstalk data to BIDS.

example usage:
$ mne_bids crosstalk_to_bids --subject_id=01 --session=test
--bids_root=bids_root --file=ct_sparse.fif

"""
# Authors: Richard Höchenberger <richard.hoechenberger@gmail.com>
#
# License: BSD-3-Clause

from mne.utils import logger

import mne_bids
from mne_bids import BIDSPath, write_meg_crosstalk


def run():
    """Run the crosstalk_to_bids command."""
    from mne.commands.utils import get_optparser

    parser = get_optparser(__file__, usage="usage: %prog options args",
                           prog_prefix='mne_bids',
                           version=mne_bids.__version__)

    parser.add_option('--bids_root', dest='bids_root',
                      help='The path of the folder containing the BIDS '
                           'dataset')
    parser.add_option('--subject_id', dest='subject',
                      help=('Subject name'))
    parser.add_option('--session_id', dest='session',
                      help='Session name')
    parser.add_option('--file', dest='fname',
                      help='The path of the crosstalk file')
    parser.add_option('--verbose', dest='verbose', action='store_true',
                      help='Whether do generate additional diagnostic output')

    opt, args = parser.parse_args()
    if args:
        parser.print_help()
        parser.error(f'Please do not specify arguments without flags. '
                     f'Got: {args}.\n')

    if opt.bids_root is None:
        parser.print_help()
        parser.error('You must specify bids_root')
    if opt.subject is None:
        parser.print_help()
        parser.error('You must specify a subject')

    bids_path = BIDSPath(subject=opt.subject, session=opt.session,
                         root=opt.bids_root)

    logger.info(f'Writing crosstalk file {bids_path.basename} …')
    write_meg_crosstalk(fname=opt.fname, bids_path=bids_path,
                        verbose=opt.verbose)


if __name__ == '__main__':
    run()



"""
=========================================================
03. Interactive data inspection and bad channel selection
=========================================================

You can use MNE-BIDS interactively inspect your  MEG or (i)EEG data.
Problematic channels can be marked as "bad", for example if the connected
sensor produced mostly noise – or no signal at all. Similarly, you can declare
channels as "good", should you discover they were incorrectly marked as bad.
Bad channel selection can also be performed non-interactively.

Furthermore, you can view and edit the experimental events and mark time
segments as "bad".

.. _MNE-Python Annotations tutorial: https://mne.tools/stable/auto_tutorials/raw/30_annotate_raw.html#annotating-raw-objects-interactively
"""  # noqa: E501 D400 D205

# Authors: Richard Höchenberger <richard.hoechenberger@gmail.com>
#
# License: BSD-3-Clause

# %%
# We will demonstrate how to mark individual channels as bad on the MNE
# "sample" dataset. After that, we will mark channels as good again.
#
# Let's start by importing the required modules and functions, reading the
# "sample" data, and writing it in the BIDS format.

import os.path as op
import shutil

import mne
from mne_bids import (BIDSPath, write_raw_bids, read_raw_bids,
                      inspect_dataset, mark_channels)

data_path = mne.datasets.sample.data_path()
raw_fname = op.join(data_path, 'MEG', 'sample', 'sample_audvis_raw.fif')
events_fname = op.join(data_path, 'MEG', 'sample', 'sample_audvis_raw-eve.fif')
event_id = {'Auditory/Left': 1, 'Auditory/Right': 2, 'Visual/Left': 3,
            'Visual/Right': 4, 'Smiley': 5, 'Button': 32}
bids_root = op.join(data_path, '..', 'MNE-sample-data-bids')
bids_path = BIDSPath(subject='01', session='01', task='audiovisual', run='01',
                     root=bids_root)

# %%
# To ensure the output path doesn't contain any leftover files from previous
# tests and example runs, we simply delete it.
#
# .. warning:: Do not delete directories that may contain important data!
#

if op.exists(bids_root):
    shutil.rmtree(bids_root)

# %%
# Now write the raw data to BIDS.

raw = mne.io.read_raw_fif(raw_fname, verbose=False)
raw.info['line_freq'] = 60  # Specify power line frequency as required by BIDS.
write_raw_bids(raw, bids_path=bids_path, events=events_fname,
               event_id=event_id, overwrite=True, verbose=False)

# %%
# Interactive use
# ---------------
#
# Using :func:`mne_bids.inspect_dataset`, we can interactively explore the raw
# data and toggle the channel status – ``bad`` or ``good`` – by clicking on the
# respective traces or channel names. If there are any SSP projectors stored
# with the data, a small popup window will allow you to toggle the projectors
# on and off. If you changed the selection of bad channels, you will be
# prompted whether you would like to save the changes when closing the main
# window. Your raw data and the `*_channels.tsv` sidecar file will be updated
# upon saving.

inspect_dataset(bids_path)

# %%
# You can even apply frequency filters when viewing the data: A high-pass
# filter can remove slow drifts, while a low-pass filter will get rid of
# high-frequency artifacts. This can make visual inspection easier. Let's
# apply filters with a 1-Hz high-pass cutoff, and a 30-Hz low-pass cutoff:

inspect_dataset(bids_path, l_freq=1., h_freq=30.)

# %%
# By pressing the ``A`` key, you can toggle annotation mode to add, edit, or
# remove experimental events, or to mark entire time periods as bad. Please see
# the `MNE-Python Annotations tutorial`_ for an introduction to the interactive
# interface. If you're closing the main window after changing the annotations,
# you will be prompted whether you wish to save the changes. Your raw data and
# the `*_events.tsv` sidecar file will be updated upon saving.
#
# Non-interactive (programmatic) bad channel selection
# ----------------------------------------------------
#
# Read the (now BIDS-formatted) data and print a list of channels currently
# marked as bad.

raw = read_raw_bids(bids_path=bids_path, verbose=False)
print(f'The following channels are currently marked as bad:\n'
      f'    {", ".join(raw.info["bads"])}\n')

# %%
# So currently, two channels are marked as bad: ``EEG 053`` and ``MEG 2443``.
# Let's assume that through visual data inspection, we found that two more
# MEG channels are problematic, and we would like to mark them as bad as well.
# To do that, we simply add them to a list, which we then pass to
# :func:`mne_bids.mark_channels`:

bads = ['MEG 0112', 'MEG 0131']
mark_channels(bids_path=bids_path, ch_names=bads, status='bad',
              verbose=False)

# %%
# That's it! Let's verify the result.

raw = read_raw_bids(bids_path=bids_path, verbose=False)
print(f'After marking MEG 0112 and MEG 0131 as bad, the following channels '
      f'are now marked as bad:\n    {", ".join(raw.info["bads"])}\n')

# %%
# As you can see, now a total of **four** channels is marked as bad: the ones
# that were already bad when we started – ``EEG 053`` and ``MEG 2443`` – and
# the two channels we passed to :func:`mne_bids.mark_channels` –
# ``MEG 0112`` and ``MEG 0131``. This shows that marking bad channels via
# :func:`mne_bids.mark_channels`, by default, is an **additive** procedure,
# which allows you to mark additional channels as bad while retaining the
# information about all channels that had *previously* been marked as bad.
#
# If you instead would like to **replace** the collection of bad channels
# entirely, pass the argument ``overwrite=True``:

bads = ['MEG 0112', 'MEG 0131']
mark_channels(bids_path=bids_path, ch_names=bads, status='bad', verbose=False)

raw = read_raw_bids(bids_path=bids_path, verbose=False)
print(f'After marking MEG 0112 and MEG 0131 as bad and passing '
      f'`overwrite=True`, the following channels '
      f'are now marked as bad:\n    {", ".join(raw.info["bads"])}\n')

# %%
# Lastly, if you're looking for a way to mark all channels as good, simply
# pass an empty list as ``ch_names``, combined with ``overwrite=True``:

bads = []
mark_channels(bids_path=bids_path, ch_names=bads, status='bad', verbose=False)

raw = read_raw_bids(bids_path=bids_path, verbose=False)
print(f'After passing `ch_names=[]` and `overwrite=True`, the following '
      f'channels are now marked as bad:\n    {", ".join(raw.info["bads"])}\n')



"""Custom sphinx extension to generate docs for the command line interface.

Inspired by MNE-Python's `gen_commands.py`
see: github.com/mne-tools/mne-python/blob/main/doc/sphinxext/gen_commands.py
"""
# Authors: Eric Larson <larson.eric.d@gmail.com>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#
# License: BSD-3-Clause
import os
import glob
from os import path as op
import subprocess
import sys

import sphinx.util
from mne.utils import run_subprocess, _replace_md5


def setup(app):
    """Set up the app."""
    app.connect('builder-inited', generate_cli_rst)


# Header markings go:
# 1. =/= : Page title
# 2. =   : Command name
# 3. -/- : Command description
# 4. -   : Command sections (Examples, Notes)

header = """\
:orphan:

.. _python_cli:

=====================================
MNE-BIDS Command Line Interface (CLI)
=====================================

Here we list the MNE-BIDS tools that you can use from the command line.

"""

command_rst = """

.. _gen_%s:

%s
%s

.. rst-class:: callout

%s

"""


def generate_cli_rst(app=None):
    """Generate the command line interface docs."""
    out_dir = op.abspath(op.join(op.dirname(__file__), '..', 'generated'))
    if not op.isdir(out_dir):
        os.mkdir(out_dir)
    out_fname = op.join(out_dir, 'cli.rst.new')

    cli_path = op.abspath(
        op.join(os.path.dirname(__file__), '..', '..', 'mne_bids', 'commands'))
    fnames = sorted([
        op.basename(fname)
        for fname in glob.glob(op.join(cli_path, 'mne_bids*.py'))])
    iterator = sphinx.util.status_iterator(
        fnames, 'generating MNE-BIDS cli help ... ', length=len(fnames))
    with open(out_fname, 'w', encoding='utf-8') as f:
        f.write(header)
        for fname in iterator:
            cmd_name = fname[:-3]
            run_name = op.join(cli_path, fname)
            output, _ = run_subprocess([sys.executable, run_name, '--help'],
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE, verbose=False)
            output = output.splitlines()

            # Swap usage and title lines
            output[0], output[2] = output[2], output[0]

            # Add header marking
            for idx in (1, 0):
                output.insert(idx, '-' * len(output[0]))

            # Add code styling for the "Usage: " line
            for li, line in enumerate(output):
                if line.startswith('Usage: mne_bids '):
                    output[li] = 'Usage: ``%s``' % line[7:]
                    break

            # Turn "Options:" into field list
            if 'Options:' in output:
                ii = output.index('Options:')
                output[ii] = 'Options'
                output.insert(ii + 1, '-------')
                output.insert(ii + 2, '')
                output.insert(ii + 3, '.. rst-class:: field-list cmd-list')
                output.insert(ii + 4, '')
            output = '\n'.join(output)
            f.write(command_rst % (cmd_name,
                                   cmd_name.replace('mne_bids_', 'mne_bids '),
                                   '=' * len(cmd_name),
                                   output))
    _replace_md5(out_fname)
    print('[Done]')


# This is useful for testing/iterating to see what the result looks like
if __name__ == '__main__':
    generate_cli_rst()



"""
.. _read_bids_datasets-example:

======================
01. Read BIDS datasets
======================

When working with electrophysiological data in the BIDS format, an important
resource is the `OpenNeuro <https://openneuro.org/>`_ database. OpenNeuro
works great with MNE-BIDS because every dataset must pass a validator
that tests to ensure its format meets BIDS specifications before the dataset
can be uploaded, so you know the data will work with a script like in this
example without modification.

We have various data types that can be loaded via the ``read_raw_bids``
function:

- MEG
- EEG (scalp electrodes)
- iEEG (ECoG and SEEG)
- the anatomical MRI scan of a study participant

In this tutorial, we show how ``read_raw_bids`` can be used to load and
inspect BIDS-formatted data.

"""
# Authors: Adam Li <adam2392@gmail.com>
#          Richard Höchenberger <richard.hoechenberger@gmail.com>
#          Alex Rockhill <aprockhill@mailbox.org>
#
# License: BSD-3-Clause

# %%
# Imports
# -------
# We are importing everything we need for this example:
import os
import os.path as op
import openneuro

from mne.datasets import sample
from mne_bids import BIDSPath, read_raw_bids, print_dir_tree, make_report

# %%
# Download a subject's data from an OpenNeuro BIDS dataset
# --------------------------------------------------------
#
# Download the data, storing each in a ``target_dir`` target directory, which,
# in ``mne-bids`` terminology, is the `root` of each BIDS dataset. This example
# uses this `EEG dataset <https://openneuro.org/datasets/ds002778>`_ of
# resting-state recordings of patients with Parkinson's disease.
#

# .. note: If the keyword argument include is left out of
#          ``openneuro.download``, the whole dataset will be downloaded.
#          We're just using data from one subject to reduce the time
#          it takes to run the example.

dataset = 'ds002778'
subject = 'pd6'

# Download one subject's data from each dataset
bids_root = op.join(op.dirname(sample.data_path()), dataset)
if not op.isdir(bids_root):
    os.makedirs(bids_root)

openneuro.download(dataset=dataset, target_dir=bids_root,
                   include=[f'sub-{subject}'])

# %%
# Explore the dataset contents
# ----------------------------
#
# We can use MNE-BIDS to print a tree of all
# included files and folders. We pass the ``max_depth`` parameter to
# `mne_bids.print_dir_tree` to the output to four levels of folders, for
# better readability in this example.

print_dir_tree(bids_root, max_depth=4)

# %%
# We can even ask MNE-BIDS to produce a human-readbale summary report
# on the dataset contents.

print(make_report(bids_root))

# %%
# Now it's time to get ready for reading some of the data! First, we need to
# create an :class:`mne_bids.BIDSPath`, which is the workhorse object of
# MNE-BIDS when it comes to file and folder operations.
#
# For now, we're interested only in the EEG data in the BIDS root directory
# of the Parkinson's disease patient dataset. There were two sessions, one
# where the patients took their regular anti-Parkinsonian medications and
# one where they abstained for more than twelve hours. Let's start with the
# off-medication session.

datatype = 'eeg'
session = 'off'
bids_path = BIDSPath(root=bids_root, session=session, datatype=datatype)

# %%
# We can now retrieve a list of all MEG-related files in the dataset:

print(bids_path.match())

# %%
# The returned list contains ``BIDSpaths`` of 3 files:
# ``sub-pd6_ses-off_task-rest_channels.tsv``,
# ``sub-pd6_ses-off_task-rest_events.tsv``, and
# ``sub-pd6_ses-off_task-rest_eeg.bdf``.
# The first two are so-called sidecar files that contain information on the
# recording channels and experimental events, and the third one is the actual
# data file.
#
# Prepare reading the data
# ------------------------
#
# There is only one subject and one experimental task (``rest``).
# Let's use this knowledge to create a new ``BIDSPath`` with
# all the information required to actually read the EEG data. We also need to
# pass a ``suffix``, which is the last part of the filename just before the
# extension -- ``'channels'`` and ``'events'`` for the two TSV files in
# our example, and ``'eeg'`` for EEG raw data. For MEG and EEG raw data, the
# suffix is identical to the datatype, so don't let yourself be confused here!

task = 'rest'
suffix = 'eeg'

bids_path = BIDSPath(subject=subject, session=session, task=task,
                     suffix=suffix, datatype=datatype, root=bids_root)

# %%
# Now let's print the contents of ``bids_path``.

print(bids_path)

# %%
# You probably noticed two things: Firstly, this looks like an ordinary string
# now, not like the more-or-less neatly formatted output we saw before. And
# secondly, that there's suddenly a filename extension which we never specified
# anywhere!
#
# The reason is that when you call ``print(bids_path)``, ``BIDSPath`` returns
# a string representation of ``BIDSPath.fpath``, which looks different. If,
# instead, you simply typed ``bids_path`` (or ``print(repr(bids_path))``, which
# is the same) into your Python console, you would get the nicely formatted
# output:

bids_path

# %%
# The ``root`` here is – you guessed it – the directory we passed via the
# ``root`` parameter: the "home" of our BIDS dataset. The ``datatype``, again,
# is self-explanatory. The ``basename``, on the other hand, is created
# automatically based on the suffix and **BIDS entities**  we passed to
# ``BIDSPath``: in our case, ``subject``, ``session`` and ``task``.
#
# .. note::
#   There are many more supported entities, the most-commonly used among them
#   probably being ``acquisition``. Please see
#   :ref:`our introduction to BIDSPath <bidspath-example>` to learn more
#   about entities, ``basename``, and ``BIDSPath`` in general.
#
# But what about that filename extension, now? ``BIDSPath.fpath``, which –
# as you hopefully remember – is invoked when you run ``print(bids_path)`` –
# employs some heuristics to auto-detect some missing filename components.
# Omitting the filename extension in your script can make your code
# more portable. Note that, however, you **can** explicitly specify an
# extension too, by passing e.g. ``extension='.bdf'`` to ``BIDSPath``.

# %%
# Read the data
# -------------
#
# Let's read the data! It's just a single line of code.

raw = read_raw_bids(bids_path=bids_path, verbose=False)

# %%
# Now we can inspect the ``raw`` object to check that it contains to correct
# metadata.
#
# Basic subject metadata is here.

print(raw.info['subject_info'])

# %%
# Power line frequency is here.

print(raw.info['line_freq'])

# %%
# Sampling frequency is here.

print(raw.info['sfreq'])

# %%
# Events are now Annotations
print(raw.annotations)

# %%
# Plot the raw data.

raw.plot()

# %%
# .. LINKS
#
# .. _parkinsons_eeg_dataset:
#    https://openneuro.org/datasets/ds002778
#



"""Make BIDS report from dataset and sidecar files."""
# Authors: Adam Li <adam2392@gmail.com>
#
# License: BSD-3-Clause
import json
import os.path as op
import textwrap
from pathlib import Path

import numpy as np
import jinja2
from mne.utils import warn, logger, verbose

from mne_bids.config import DOI, ALLOWED_DATATYPES
from mne_bids.tsv_handler import _from_tsv
from mne_bids.path import (get_bids_path_from_fname, get_datatypes,
                           get_entity_vals, BIDSPath,
                           _parse_ext, _find_matching_sidecar)


jinja_env = jinja2.Environment(
    loader=jinja2.PackageLoader(
        package_name='mne_bids.report',
        package_path='templates'
    )
)


def _pretty_str(listed):
    # make strings a sequence of ',' and 'and'
    if not isinstance(listed, list):
        listed = list(listed)

    if len(listed) <= 1:
        return ','.join(listed)
    return '{}, and {}'.format(', '.join(listed[:-1]), listed[-1])


def _range_str(minval, maxval, meanval, stdval, n_unknown, type):
    if minval == 'n/a':
        return 'ages all unknown'

    if n_unknown > 0:
        unknown_str = f'; {n_unknown} with unknown {type}'
    else:
        unknown_str = ''
    return (
        f'ages ranged from {round(minval, 2)} to {round(maxval, 2)} '
        f'(mean = {round(meanval, 2)}, std = {round(stdval, 2)}{unknown_str})'
    )


def _summarize_participant_hand(hands):
    n_unknown = len([hand for hand in hands if hand == 'n/a'])

    if n_unknown == len(hands):
        return 'handedness were all unknown'

    n_rhand = len([hand for hand in hands if hand.upper() == 'R'])
    n_lhand = len([hand for hand in hands if hand.upper() == 'L'])
    n_ambidex = len([hand for hand in hands if hand.upper() == 'A'])

    return f'comprised of {n_rhand} right hand, {n_lhand} left hand ' \
           f'and {n_ambidex} ambidextrous'


def _summarize_participant_sex(sexs):
    n_unknown = len([sex for sex in sexs if sex == 'n/a'])

    if n_unknown == len(sexs):
        return 'sex were all unknown'

    n_males = len([sex for sex in sexs if sex.upper() == 'M'])
    n_females = len([sex for sex in sexs if sex.upper() == 'F'])

    return f'comprised of {n_males} male and {n_females} female participants'


def _length_recording_str(length_recordings):
    import numpy as np
    if length_recordings is None:
        return ''

    min_record_length = round(np.min(length_recordings), 2)
    max_record_length = round(np.max(length_recordings), 2)
    mean_record_length = round(np.mean(length_recordings), 2)
    std_record_length = round(np.std(length_recordings), 2)
    total_record_length = round(sum(length_recordings), 2)

    return (
        f'Recording durations ranged from {min_record_length} to '
        f'{max_record_length} seconds '
        f'(mean = {mean_record_length}, std = {std_record_length}), '
        f'for a total of {total_record_length} seconds of data recorded '
        f'over all scans.'
    )


def _summarize_software_filters(software_filters):
    if software_filters in [{}, 'n/a']:
        return ''

    msg = ''
    for key, value in software_filters.items():
        msg += f'{key}'

        if isinstance(value, dict) and value:
            parameters = []
            for param_name, param_value in value.items():
                if param_name and param_value:
                    parameters.append(f'{param_value} {param_name}')
            if parameters:
                msg += ' with parameters '
                msg += ', '.join(parameters)
    return msg


def _pretty_dict(template_dict):
    """Remove problematic blank spaces."""
    for key, val in template_dict.items():
        if val == ' ':
            template_dict[key] = 'n/a'


def _summarize_dataset(root):
    """Summarize the dataset_desecription.json file.

    Required dataset descriptors include:
        - Name
        - BIDSVersion

    Added descriptors include:
        - Authors
        - DOI

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.

    Returns
    -------
    template_dict : dict
        A dictionary of values for various template strings.
    """
    dataset_descrip_fpath = op.join(root,
                                    'dataset_description.json')
    if not op.exists(dataset_descrip_fpath):
        return dict()

    # read file and 'REQUIRED' components of it
    with open(dataset_descrip_fpath, 'r', encoding='utf-8-sig') as fin:
        dataset_description = json.load(fin)

    # create dictionary to pass into template string
    name = dataset_description['Name']
    bids_version = dataset_description['BIDSVersion']
    authors = dataset_description['Authors']
    template_dict = {
        'name': name,
        'bids_version': bids_version,
        'mne_bids_doi': DOI,
        'authors': _pretty_str(authors),
    }
    _pretty_dict(template_dict)
    return template_dict


def _summarize_participants_tsv(root):
    """Summarize `participants.tsv` file in BIDS root directory.

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.

    Returns
    -------
    template_dict : dict
        A dictionary of values for various template strings.
    """
    participants_tsv_fpath = op.join(root, 'participants.tsv')
    if not op.exists(participants_tsv_fpath):
        return dict()

    participants_tsv = _from_tsv(str(participants_tsv_fpath))
    p_ids = participants_tsv['participant_id']
    logger.info(f'Summarizing participants.tsv {participants_tsv_fpath}...')

    # summarize sex count statistics
    keys = ['M', 'F', 'n/a']
    p_sex = participants_tsv.get('sex')
    # phrasing works for both sex and gender
    p_gender = participants_tsv.get('gender')
    sexs = ['n/a']
    if p_sex or p_gender:
        # only summarize sex if it conforms to `keys` referenced above
        p_sex = p_gender if p_sex is None else p_sex
        if all([sex.upper() in keys
                for sex in p_sex if sex != 'n/a']):
            sexs = p_sex

    # summarize hand count statistics
    keys = ['R', 'L', 'A', 'n/a']
    p_hands = participants_tsv.get('hand')
    hands = ['n/a']
    if p_hands:
        # only summarize handedness if it conforms to
        # mne-bids handedness
        if all([hand.upper() in keys
                for hand in p_hands if hand != 'n/a']):
            hands = p_hands

    # summarize age statistics: mean, std, min, max
    p_ages = participants_tsv.get('age')
    min_age, max_age = 'n/a', 'n/a'
    mean_age, std_age = 'n/a', 'n/a'
    n_age_unknown = len(p_ages) if p_ages else len(p_ids)
    if p_ages:
        # only summarize age if they are numerics
        if all([age.isnumeric() for age in p_ages if age != 'n/a']):
            age_list = [float(age) for age in p_ages if age != 'n/a']
            n_age_unknown = len(p_ids) - len(age_list)
            if age_list:
                min_age, max_age = np.min(age_list), np.max(age_list)
                mean_age, std_age = np.mean(age_list), np.std(age_list)

    template_dict = {
        'sexs': _summarize_participant_sex(sexs),
        'hands': _summarize_participant_hand(hands),
        'ages': _range_str(
            min_age, max_age, mean_age, std_age, n_age_unknown,
            'age'
        )
    }
    return template_dict


def _summarize_scans(root, session=None):
    """Summarize scans in BIDS root directory.

    Summarizes scans only if there is a *_scans.tsv file.

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.
    session : str, optional
        The session for a item. Corresponds to "ses".

    Returns
    -------
    template_dict : dict
        A dictionary of values for various template strings.

    """
    root = Path(root)
    if session is None:
        search_str = '*_scans.tsv'
    else:
        search_str = f'*ses-{session}' \
                     f'*_scans.tsv'
    scans_fpaths = list(root.rglob(search_str))
    if len(scans_fpaths) == 0:
        warn('No *scans.tsv files found. Currently, '
             'we do not generate a report without the scans.tsv files.')
        return dict()

    logger.info(f'Summarizing scans.tsv files {scans_fpaths}...')

    # summarize sidecar.json, channels.tsv template
    sidecar_dict = _summarize_sidecar_json(root, scans_fpaths)
    channels_dict = _summarize_channels_tsv(root, scans_fpaths)
    template_dict = dict()
    template_dict.update(**sidecar_dict)
    template_dict.update(**channels_dict)

    return template_dict


def _summarize_sidecar_json(root, scans_fpaths):
    """Summarize scans in BIDS root directory.

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.
    scans_fpaths : list
        A list of all *_scans.tsv files in ``root``. The summary
        will occur for all scans listed in the *_scans.tsv files.

    Returns
    -------
    template_dict : dict
        A dictionary of values for various template strings.

    """
    n_scans = 0
    powerlinefreqs, sfreqs = set(), set()
    manufacturers = set()
    length_recordings = []

    # loop through each scan
    for scan_fpath in scans_fpaths:
        # load in the scans.tsv file
        # and read metadata for each scan
        scans_tsv = _from_tsv(scan_fpath)
        scans = scans_tsv['filename']
        for scan in scans:
            # summarize metadata of recordings
            bids_path, ext = _parse_ext(scan)
            datatype = op.dirname(scan)
            if datatype not in ALLOWED_DATATYPES:
                continue

            n_scans += 1

            # convert to BIDSPath
            if not isinstance(bids_path, BIDSPath):
                bids_path = get_bids_path_from_fname(bids_path)
            bids_path.root = root

            # XXX: improve to allow emptyroom
            if bids_path.subject == 'emptyroom':
                continue

            sidecar_fname = _find_matching_sidecar(bids_path=bids_path,
                                                   suffix=datatype,
                                                   extension='.json')
            with open(sidecar_fname, 'r', encoding='utf-8-sig') as fin:
                sidecar_json = json.load(fin)

            # aggregate metadata from each scan
            # REQUIRED kwargs
            sfreq = sidecar_json['SamplingFrequency']
            powerlinefreq = str(sidecar_json['PowerLineFrequency'])
            software_filters = sidecar_json.get('SoftwareFilters')
            if not software_filters:
                software_filters = 'n/a'

            # RECOMMENDED kwargs
            manufacturer = sidecar_json.get('Manufacturer', 'n/a')
            record_duration = sidecar_json.get('RecordingDuration', 'n/a')

            sfreqs.add(str(np.round(sfreq, 2)))
            powerlinefreqs.add(str(powerlinefreq))
            if manufacturer != 'n/a':
                manufacturers.add(manufacturer)
            length_recordings.append(record_duration)

    # XXX: length summary is only allowed, if no 'n/a' was found
    if any([dur == 'n/a' for dur in length_recordings]):
        length_recordings = None

    template_dict = {
        'n_scans': n_scans,
        'manufacturer': _pretty_str(manufacturers),
        'sfreq': _pretty_str(sfreqs),
        'powerlinefreq': _pretty_str(powerlinefreqs),
        'software_filters': _summarize_software_filters(software_filters),
        'length_recordings': _length_recording_str(length_recordings),
    }
    return template_dict


def _summarize_channels_tsv(root, scans_fpaths):
    """Summarize channels.tsv data in BIDS root directory.

    Currently, summarizes all REQUIRED components of channels
    data, and some RECOMMENDED and OPTIONAL components.

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.
    scans_fpaths : list
        A list of all *_scans.tsv files in ``root``. The summary
        will occur for all scans listed in the *_scans.tsv files.

    Returns
    -------
    template_dict : dict
        A dictionary of values for various template strings.
    """
    root = Path(root)

    # keep track of channel type, status
    ch_status_count = {'bad': [], 'good': []}
    ch_count = []

    # loop through each scan
    for scan_fpath in scans_fpaths:
        # load in the scans.tsv file
        # and read metadata for each scan
        scans_tsv = _from_tsv(scan_fpath)
        scans = scans_tsv['filename']
        for scan in scans:
            # summarize metadata of recordings
            bids_path, _ = _parse_ext(scan)
            datatype = op.dirname(scan)
            if datatype not in ['meg', 'eeg', 'ieeg']:
                continue

            # convert to BIDSPath
            if not isinstance(bids_path, BIDSPath):
                bids_path = get_bids_path_from_fname(bids_path)
            bids_path.root = root

            # XXX: improve to allow emptyroom
            if bids_path.subject == 'emptyroom':
                continue

            channels_fname = _find_matching_sidecar(bids_path=bids_path,
                                                    suffix='channels',
                                                    extension='.tsv')

            # summarize channels.tsv
            channels_tsv = _from_tsv(channels_fname)
            for status in ch_status_count.keys():
                ch_status = [ch for ch in channels_tsv['status']
                             if ch == status]
                ch_status_count[status].append(len(ch_status))
            ch_count.append(len(channels_tsv['name']))

    # create summary template strings for status
    template_dict = {
        'mean_chs': np.mean(ch_count),
        'std_chs': np.std(ch_count),
        'mean_good_chs': np.mean(ch_status_count['good']),
        'std_good_chs': np.std(ch_status_count['good']),
        'mean_bad_chs': np.mean(ch_status_count['bad']),
        'std_bad_chs': np.std(ch_status_count['bad']),
    }
    for key, val in template_dict.items():
        template_dict[key] = round(val, 2)
    return template_dict


@verbose
def make_report(root, session=None, verbose=None):
    """Create a methods paragraph string from BIDS dataset.

    Summarizes the REQUIRED components in the BIDS specification
    and also some RECOMMENDED components. Currently, the methods
    paragraph summarize the:

      - dataset_description.json file
      - (optional) participants.tsv file
      - (optional) datatype-agnostic files for (M/I)EEG data,
        which reads files from the ``*_scans.tsv`` file.

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.
    session : str | None
            The (optional) session for a item. Corresponds to "ses".
    %(verbose)s

    Returns
    -------
    paragraph : str
        The paragraph wrapped with 80 characters per line
        describing the summary of the subjects.
    """
    # high level summary
    subjects = get_entity_vals(root, entity_key='subject')
    sessions = get_entity_vals(root, entity_key='session')
    modalities = get_datatypes(root)

    # only summarize allowed modalities (MEG/EEG/iEEG) data
    # map them to a pretty looking string
    datatype_map = {
        'meg': 'MEG',
        'eeg': 'EEG',
        'ieeg': 'iEEG',
    }
    modalities = [datatype_map[datatype] for datatype in modalities
                  if datatype in datatype_map.keys()]

    # REQUIRED: dataset_description.json summary
    dataset_summary = _summarize_dataset(root)

    # RECOMMENDED: participants summary
    participant_summary = _summarize_participants_tsv(root)

    # RECOMMENDED: scans summary
    scans_summary = _summarize_scans(root, session=session)

    dataset_agnostic_summary = scans_summary.copy()
    dataset_agnostic_summary['system'] = _pretty_str(modalities)

    # turn off 'recommended' report summary
    # if files are not available to summarize
    if not participant_summary:
        participants_info = ''
    else:
        particpants_info_template = jinja_env.get_template(
            'participants.jinja'
        )
        participants_info = particpants_info_template.render(
            **participant_summary
        )
        logger.info(f'The participant template found: {participants_info}')

    if not scans_summary:
        datatype_agnostic_info = ''
    else:
        datatype_agnostic_template = jinja_env.get_template(
            'datatype_agnostic.jinja'
        )
        datatype_agnostic_info = datatype_agnostic_template.render(
            **dataset_agnostic_summary
        )

    dataset_summary.update({
        'n_subjects': len(subjects),
        'participants_info': participants_info,
        'n_sessions': len(sessions),
        'sessions': _pretty_str(sessions),
    })

    # XXX: add channel summary for modalities (ieeg, meg, eeg)
    # create the content and mne Template
    # lower-case templates are "Recommended",
    # while upper-case templates are "Required".

    dataset_summary_template = jinja_env.get_template('dataset_summary.jinja')
    dataset_summary_info = dataset_summary_template.render(**dataset_summary)

    # Concatenate info and clean the paragraph
    paragraph = f'{dataset_summary_info}\n{datatype_agnostic_info}'
    paragraph = paragraph.replace('\n', ' ')
    while '  ' in paragraph:
        paragraph = paragraph.replace('  ', ' ')

    return '\n'.join(textwrap.wrap(paragraph, width=80))



"""Test the interactive data inspector."""

import os.path as op
import pytest
from functools import partial

import numpy as np
from matplotlib.backend_bases import KeyEvent

import mne
from mne.datasets import testing
from mne.utils._testing import requires_module
from mne.viz.utils import _fake_click

from mne_bids import (BIDSPath, read_raw_bids, write_raw_bids, inspect_dataset,
                      write_meg_calibration, write_meg_crosstalk)
import mne_bids.inspect
from mne_bids.read import _from_tsv

from test_read import warning_str

requires_matplotlib = partial(requires_module, name='matplotlib',
                              call='import matplotlib')

_bids_path = BIDSPath(subject='01', session='01', run='01', task='testing',
                      datatype='meg')
data_path = testing.data_path(download=False)


def setup_bids_test_dir(bids_root):
    """Return path to a written test BIDS dir."""
    raw_fname = op.join(data_path, 'MEG', 'sample',
                        'sample_audvis_trunc_raw.fif')

    event_id = {'Auditory/Left': 1, 'Auditory/Right': 2, 'Visual/Left': 3,
                'Visual/Right': 4, 'Smiley': 5, 'Button': 32}
    events_fname = op.join(data_path, 'MEG', 'sample',
                           'sample_audvis_trunc_raw-eve.fif')
    cal_fname = op.join(data_path, 'SSS', 'sss_cal_mgh.dat')
    crosstalk_fname = op.join(data_path, 'SSS', 'ct_sparse.fif')

    raw = mne.io.read_raw_fif(raw_fname)
    raw.info['line_freq'] = 60

    # Drop unknown events.
    events = mne.read_events(events_fname)
    events = events[events[:, 2] != 0]

    bids_path = _bids_path.copy().update(root=bids_root)
    write_raw_bids(raw, bids_path=bids_path, events=events,
                   event_id=event_id, overwrite=True)
    write_meg_calibration(cal_fname, bids_path=bids_path)
    write_meg_crosstalk(crosstalk_fname, bids_path=bids_path)

    return bids_root


@requires_matplotlib
@testing.requires_testing_data
@pytest.mark.parametrize('save_changes', (True, False))
@pytest.mark.filterwarnings(warning_str['channel_unit_changed'])
def test_inspect_single_file(tmp_path, save_changes):
    """Test inspecting a dataset consisting of only a single file."""
    from mne.utils._testing import _click_ch_name
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('Agg')
    plt.close('all')

    bids_root = setup_bids_test_dir(tmp_path)
    bids_path = _bids_path.copy().update(root=bids_root)
    raw = read_raw_bids(bids_path=bids_path, verbose='error')
    old_bads = raw.info['bads'].copy()

    inspect_dataset(bids_path)
    raw_fig = mne_bids.inspect._global_vars['raw_fig']

    # Mark some channels as bad by clicking on their name.
    _click_ch_name(raw_fig, ch_index=0, button=1)
    _click_ch_name(raw_fig, ch_index=1, button=1)
    _click_ch_name(raw_fig, ch_index=4, button=1)

    # Closing the window should open a dialog box.
    key_event = KeyEvent(
        name='Close', canvas=raw_fig.canvas, key=raw_fig.mne.close_key
    )
    raw_fig.canvas.callbacks.process('key_press_event', key_event)
    fig_dialog = mne_bids.inspect._global_vars['dialog_fig']

    if save_changes:
        key = 'return'
    else:
        key = 'escape'
    key_event = KeyEvent(name='Save', canvas=fig_dialog.canvas, key=key)
    fig_dialog.canvas.callbacks.process('key_press_event', key_event)

    raw = read_raw_bids(bids_path=bids_path, verbose='error')
    new_bads = raw.info['bads'].copy()

    if save_changes:
        assert len(new_bads) > len(old_bads)
    else:
        assert old_bads == new_bads


@requires_matplotlib
@testing.requires_testing_data
@pytest.mark.filterwarnings(warning_str['channel_unit_changed'])
def test_inspect_multiple_files(tmp_path):
    """Test inspecting a dataset consisting of more than one file."""
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('Agg')
    plt.close('all')

    bids_root = setup_bids_test_dir(tmp_path)
    bids_path = _bids_path.copy().update(root=bids_root)

    # Create a second subject
    raw = read_raw_bids(bids_path=bids_path, verbose='error')
    write_raw_bids(raw, bids_path.copy().update(subject='02'))
    del raw

    # Inspection should end with the second subject.
    inspect_dataset(bids_path.copy().update(subject=None))
    raw_fig = mne_bids.inspect._global_vars['raw_fig']
    assert raw_fig.mne.info['subject_info']['his_id'] == 'sub-02'

    key_event = KeyEvent(
        name='Close', canvas=raw_fig.canvas, key=raw_fig.mne.close_key
    )
    raw_fig.canvas.callbacks.process('key_press_event', key_event)


@requires_matplotlib
@testing.requires_testing_data
@pytest.mark.filterwarnings(warning_str['channel_unit_changed'])
def test_inspect_set_and_unset_bads(tmp_path):
    """Test marking channels as bad and later marking them as good again."""
    from mne.utils._testing import _click_ch_name
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('Agg')
    plt.close('all')

    bids_root = setup_bids_test_dir(tmp_path)
    bids_path = _bids_path.copy().update(root=bids_root)
    raw = read_raw_bids(bids_path=bids_path, verbose='error')
    orig_bads = raw.info['bads'].copy()

    # Mark some channels as bad by clicking on their name.
    inspect_dataset(bids_path, find_flat=False)
    raw_fig = mne_bids.inspect._global_vars['raw_fig']
    _click_ch_name(raw_fig, ch_index=0, button=1)
    _click_ch_name(raw_fig, ch_index=1, button=1)
    _click_ch_name(raw_fig, ch_index=4, button=1)

    # Close window and save changes.
    key_event = KeyEvent(
        name='Close', canvas=raw_fig.canvas, key=raw_fig.mne.close_key
    )
    raw_fig.canvas.callbacks.process('key_press_event', key_event)

    fig_dialog = mne_bids.inspect._global_vars['dialog_fig']
    key_event = KeyEvent(name='Save', canvas=fig_dialog.canvas, key='return')
    fig_dialog.canvas.callbacks.process('key_press_event', key_event)

    # Inspect the data again, click on two of the bad channels to mark them as
    # good.
    inspect_dataset(bids_path, find_flat=False)
    raw_fig = mne_bids.inspect._global_vars['raw_fig']
    _click_ch_name(raw_fig, ch_index=1, button=1)
    _click_ch_name(raw_fig, ch_index=4, button=1)

    # Close window and save changes.
    key_event = KeyEvent(
        name='Close', canvas=raw_fig.canvas, key=raw_fig.mne.close_key
    )
    raw_fig.canvas.callbacks.process('key_press_event', key_event)

    fig_dialog = mne_bids.inspect._global_vars['dialog_fig']

    key_event = KeyEvent(name='Save', canvas=fig_dialog.canvas, key='return')
    fig_dialog.canvas.callbacks.process('key_press_event', key_event)

    # Check marking the channels as good has actually worked.
    expected_bads = orig_bads + ['MEG 0113']
    raw = read_raw_bids(bids_path=bids_path, verbose='error')
    new_bads = raw.info['bads']
    assert set(new_bads) == set(expected_bads)


def _add_annotation(raw_fig):
    """Add an Annotation to a Raw plot."""
    data_ax = raw_fig.mne.ax_main

    key_event = KeyEvent(name='Annotation', canvas=raw_fig.canvas, key='a')
    raw_fig.canvas.callbacks.process('key_press_event', key_event)

    ann_fig = raw_fig.mne.fig_annotation
    for key in 'test':  # Annotation will be named: BAD_test
        key_event = KeyEvent(name='Bad', canvas=ann_fig.canvas, key=key)
        ann_fig.canvas.callbacks.process('key_press_event', key_event)

    key_event = KeyEvent(name='Enter', canvas=ann_fig.canvas, key='enter')
    ann_fig.canvas.callbacks.process('key_press_event', key_event)

    # Draw a 4 second long Annotation.
    _fake_click(raw_fig, data_ax, [1., 1.], xform='data', button=1,
                kind='press')
    _fake_click(raw_fig, data_ax, [5., 1.], xform='data', button=1,
                kind='motion')
    _fake_click(raw_fig, data_ax, [5., 1.], xform='data', button=1,
                kind='release')


@requires_matplotlib
@testing.requires_testing_data
@pytest.mark.filterwarnings(warning_str['channel_unit_changed'])
def test_inspect_annotations(tmp_path):
    """Test inspection of Annotations."""
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('Agg')
    plt.close('all')

    bids_root = setup_bids_test_dir(tmp_path)
    bids_path = _bids_path.copy().update(root=bids_root)
    raw = read_raw_bids(bids_path=bids_path, verbose='error')
    orig_annotations = raw.annotations.copy()

    inspect_dataset(bids_path, find_flat=False)
    raw_fig = mne_bids.inspect._global_vars['raw_fig']
    _add_annotation(raw_fig)

    # Close window and save changes.
    key_event = KeyEvent(
        name='Close', canvas=raw_fig.canvas, key=raw_fig.mne.close_key
    )
    raw_fig.canvas.callbacks.process('key_press_event', key_event)

    fig_dialog = mne_bids.inspect._global_vars['dialog_fig']
    key_event = KeyEvent(name='Save', canvas=fig_dialog.canvas, key='return')
    fig_dialog.canvas.callbacks.process('key_press_event', key_event)

    # Ensure changes were saved.
    raw = read_raw_bids(bids_path=bids_path, verbose='error')
    assert 'BAD_test' in raw.annotations.description
    annot_idx = raw.annotations.description == 'BAD_test'
    assert raw.annotations.duration[annot_idx].squeeze() == 4

    # Remove the Annotation.
    inspect_dataset(bids_path, find_flat=False)
    raw_fig = mne_bids.inspect._global_vars['raw_fig']
    data_ax = raw_fig.mne.ax_main

    key_event = KeyEvent(name='Annotations', canvas=raw_fig.canvas, key='a')
    raw_fig.canvas.callbacks.process('key_press_event', key_event)
    _fake_click(raw_fig, data_ax, [1., 1.], xform='data', button=3,
                kind='press')

    # Close window and save changes.
    key_event = KeyEvent(
        name='Close', canvas=raw_fig.canvas, key=raw_fig.mne.close_key
    )
    raw_fig.canvas.callbacks.process('key_press_event', key_event)

    fig_dialog = mne_bids.inspect._global_vars['dialog_fig']
    key_event = KeyEvent(name='Save', canvas=fig_dialog.canvas, key='return')
    fig_dialog.canvas.callbacks.process('key_press_event', key_event)

    # Ensure changes were saved.
    raw = read_raw_bids(bids_path=bids_path, verbose='error')
    assert 'BAD_test' not in raw.annotations.description
    assert raw.annotations == orig_annotations


@requires_matplotlib
@testing.requires_testing_data
@pytest.mark.filterwarnings(warning_str['channel_unit_changed'])
def test_inspect_annotations_remove_all(tmp_path):
    """Test behavior if all Annotations are removed by the user."""
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('Agg')
    plt.close('all')

    bids_root = setup_bids_test_dir(tmp_path)
    bids_path = _bids_path.copy().update(root=bids_root)
    events_tsv_fpath = (bids_path.copy()
                        .update(suffix='events', extension='.tsv')
                        .fpath)

    # Remove all Annotations.
    raw = read_raw_bids(bids_path=bids_path, verbose='error')
    raw.set_annotations(None)
    raw.load_data()
    raw.save(raw.filenames[0], overwrite=True)
    # Delete events.tsv sidecar.
    (bids_path.copy()
     .update(suffix='events', extension='.tsv')
     .fpath
     .unlink())

    # Add custom Annotation.
    inspect_dataset(bids_path, find_flat=False)
    raw_fig = mne_bids.inspect._global_vars['raw_fig']
    _add_annotation(raw_fig)

    # Close window and save changes.
    key_event = KeyEvent(
        name='Close', canvas=raw_fig.canvas, key=raw_fig.mne.close_key
    )
    raw_fig.canvas.callbacks.process('key_press_event', key_event)

    fig_dialog = mne_bids.inspect._global_vars['dialog_fig']
    key_event = KeyEvent(name='Save', canvas=fig_dialog.canvas, key='return')
    fig_dialog.canvas.callbacks.process('key_press_event', key_event)

    # events.tsv sidecar should have been created.
    assert events_tsv_fpath.exists()

    # Remove the Annotation.
    inspect_dataset(bids_path, find_flat=False)
    raw_fig = mne_bids.inspect._global_vars['raw_fig']
    data_ax = raw_fig.mne.ax_main

    key_event = KeyEvent(name='Annotations', canvas=raw_fig.canvas, key='a')
    raw_fig.canvas.callbacks.process('key_press_event', key_event)
    _fake_click(raw_fig, data_ax, [1., 1.], xform='data', button=3,
                kind='press')

    # Close window and save changes.
    key_event = KeyEvent(
        name='Close', canvas=raw_fig.canvas, key=raw_fig.mne.close_key
    )
    raw_fig.canvas.callbacks.process('key_press_event', key_event)

    fig_dialog = mne_bids.inspect._global_vars['dialog_fig']
    key_event = KeyEvent(name='Save', canvas=fig_dialog.canvas, key='return')
    fig_dialog.canvas.callbacks.process('key_press_event', key_event)

    # events.tsv sidecar should not exist anymore.
    assert not events_tsv_fpath.exists()


@requires_matplotlib
@testing.requires_testing_data
@pytest.mark.filterwarnings(warning_str['channel_unit_changed'])
def test_inspect_dont_show_annotations(tmp_path):
    """Test if show_annotations=False works."""
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('Agg')
    plt.close('all')

    bids_root = setup_bids_test_dir(tmp_path)
    bids_path = _bids_path.copy().update(root=bids_root)
    inspect_dataset(bids_path, find_flat=False, show_annotations=False)
    raw_fig = mne_bids.inspect._global_vars['raw_fig']
    assert not raw_fig.mne.annotations


@requires_matplotlib
@testing.requires_testing_data
@pytest.mark.filterwarnings(warning_str['channel_unit_changed'])
def test_inspect_bads_and_annotations(tmp_path):
    """Test adding bads and Annotations in one go."""
    from mne.utils._testing import _click_ch_name
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('Agg')
    plt.close('all')

    bids_root = setup_bids_test_dir(tmp_path)
    bids_path = _bids_path.copy().update(root=bids_root)
    raw = read_raw_bids(bids_path=bids_path, verbose='error')
    orig_bads = raw.info['bads'].copy()

    inspect_dataset(bids_path, find_flat=False)
    raw_fig = mne_bids.inspect._global_vars['raw_fig']

    # Mark some channels as bad by clicking on their name.
    _click_ch_name(raw_fig, ch_index=0, button=1)

    # Add custom Annotation.
    _add_annotation(raw_fig)

    # Close window and save changes.
    key_event = KeyEvent(
        name='Close', canvas=raw_fig.canvas, key=raw_fig.mne.close_key
    )
    raw_fig.canvas.callbacks.process('key_press_event', key_event)

    fig_dialog = mne_bids.inspect._global_vars['dialog_fig']
    key_event = KeyEvent(name='Save', canvas=fig_dialog.canvas, key='return')
    fig_dialog.canvas.callbacks.process('key_press_event', key_event)

    # Check that the changes were saved.
    raw = read_raw_bids(bids_path=bids_path, verbose='error')
    new_bads = raw.info['bads']
    expected_bads = orig_bads + ['MEG 0113']
    assert set(new_bads) == set(expected_bads)
    assert 'BAD_test' in raw.annotations.description


@requires_matplotlib
@testing.requires_testing_data
@pytest.mark.parametrize('save_changes', (True, False))
@pytest.mark.filterwarnings(warning_str['channel_unit_changed'])
def test_inspect_auto_flats(tmp_path, save_changes):
    """Test flat channel & segment detection."""
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('Agg')
    plt.close('all')

    bids_root = setup_bids_test_dir(tmp_path)
    bids_path = _bids_path.copy().update(root=bids_root)
    channels_tsv_fname = bids_path.copy().update(suffix='channels',
                                                 extension='.tsv')

    raw = read_raw_bids(bids_path=bids_path, verbose='error')

    # Inject an entirely flat channel.
    raw.load_data()
    raw._data[10] = np.zeros_like(raw._data[10], dtype=raw._data.dtype)
    # Add a a flat time segment (approx. 100 ms) to another channel
    raw._data[20, 500:500 + int(np.ceil(0.1 * raw.info['sfreq']))] = 0
    raw.save(raw.filenames[0], overwrite=True)
    old_bads = raw.info['bads'].copy()

    inspect_dataset(bids_path)
    raw_fig = mne_bids.inspect._global_vars['raw_fig']

    # Closing the window should open a dialog box.
    key_event = KeyEvent(
        name='Close', canvas=raw_fig.canvas, key=raw_fig.mne.close_key
    )
    raw_fig.canvas.callbacks.process('key_press_event', key_event)
    fig_dialog = mne_bids.inspect._global_vars['dialog_fig']

    if save_changes:
        key = 'return'
    else:
        key = 'escape'
    key_event = KeyEvent(name='Close', canvas=fig_dialog.canvas, key=key)
    fig_dialog.canvas.callbacks.process('key_press_event', key_event)

    raw = read_raw_bids(bids_path=bids_path, verbose='error')

    if save_changes:
        assert old_bads != raw.info['bads']
        assert raw.ch_names[10] in raw.info['bads']
        channels_tsv_data = _from_tsv(channels_tsv_fname)
        assert (channels_tsv_data['status_description'][10] ==
                'Flat channel, auto-detected via MNE-BIDS')
        # This channel should not have been added to `bads`, but produced a
        # flat annotation.
        assert raw.ch_names[20] not in raw.info['bads']
        assert 'BAD_flat' in raw.annotations.description
    else:
        assert old_bads == raw.info['bads']
        assert 'BAD_flat' not in raw.annotations.description


@requires_matplotlib
@testing.requires_testing_data
@pytest.mark.parametrize(('l_freq', 'h_freq'),
                         [(None, None),
                          (1, None),
                          (None, 30),
                          (1, 30)])
@pytest.mark.filterwarnings(warning_str['channel_unit_changed'])
def test_inspect_freq_filter(tmp_path, l_freq, h_freq):
    """Test frequency filter for Raw display."""
    bids_root = setup_bids_test_dir(tmp_path)
    bids_path = _bids_path.copy().update(root=bids_root)
    inspect_dataset(bids_path, l_freq=l_freq, h_freq=h_freq, find_flat=False)



"""
.. _bidspath-example:

===============================
10. An introduction to BIDSPath
===============================

BIDSPath is MNE-BIDS's working horse when it comes to file and folder
operations. Learn here how to use it.
"""
# Author: Richard Höchenberger <richard.hoechenberger@gmail.com>
#
# License: BSD-3-Clause

# %%
# Obviously, to start exploring BIDSPath, we first need to import it.

from pathlib import Path

import mne_bids
from mne_bids import BIDSPath

# %%
# Now let's discuss a little bit of background on the BIDS file and folder
# naming scheme. The first term we are going to introduce is the **BIDS root**.
# The BIDS root is simply the root folder of your BIDS dataset. For
# example, if the BIDS data of one of your studies is stored in
# `/Users/me/Studies/Study_01`, then this will be the BIDS root.
#
# Similarly, if you have **no** BIDS dataset to begin with, you need to
# consider where to store your data upon BIDS conversion. Again, the intended
# target folder will be the BIDS root of your data.
#
# For the purpose of this demonstration, let's pick the ``tiny_bids`` example
# dataset that ships with the MNE-BIDS test suite.

# We are using a pathlib.Path object for convenience, but you could just use
# a string to specify ``bids_root`` here.
bids_root = Path(mne_bids.__file__).parent / 'tests' / 'data' / 'tiny_bids'

# %%
# This refers to a folder named ``my_bids_root`` in the current working
# directory. Finally, let is create a ``BIDSPath``, and tell it about our
# BIDS root. We can then also query the ``BIDSPath`` for its root.

bids_path = BIDSPath(root=bids_root)
print(bids_path.root)

# %%
# Great! But not really useful so far. BIDS also asks us to specify **subject
# identifiers**. We can either create a new ``BIDSPath``, or update our
# existing one. The value can be retrieved via the ``.subject`` attribute.

subject = '01'

# Option 1: Create an entirely new BIDSPath.
bids_path_new = BIDSPath(subject=subject, root=bids_root)
print(bids_path_new.subject)

# Option 2: Update the existing BIDSPath in-place.
bids_path.update(subject=subject)
print(bids_path.subject)

# %%
# In this example, we are going to update the existing ``BIDSPath`` using its
# ``update()`` method. But note that all parameters we pass to this method can
# also be used when creating a ``BIDSPath``.
#
# Many studies consist of multiple **sessions**. As you may have guessed,
# BIDS specifies how to store data for each session, and consequently,
# ``BIDSPath`` handles this for you too! Let's update our ``BIDSPath`` with
# information on our experimental session, and try to retrieve it again via
# ``.session``.

session = 'eeg'
bids_path.update(session=session)
print(bids_path.session)

# %%
# Now that was easy! We're almost there! We also need to specify a
# **data type**, i.e., ``meg`` for MEG data, ``eeg`` and ``ieeg`` for EEG and
# iEEG data, or ``anat`` for anatomical MRI scans. Typically, MNE-BIDS will
# infer the data type of your data automatically, for example when writing data
# using `mne_bids.write_raw_bids`. For the sake of this example, however, we
# are going to specify the data type explicitly.

datatype = 'eeg'
bids_path.update(datatype=datatype)
print(bids_path.datatype)

# %%
# Excellent! Let's have a look at the path we have constructed!
print(bids_path)

# %%
# As you can see, ``BIDSPath`` automatically arranged all the information we
# provided such that it creates a valid BIDS folder structure. You can also
# retrieve a `pathlib.Path` object of this path:

pathlib_path = bids_path.fpath
pathlib_path

# %%
# Let's have a closer look at the components of our ``BIDSPath`` again.

bids_path

# %%
# The most interesting thing here is probably the **basename**. It's what
# MNE-BIDS uses to name individual files. The basename consists of a set of
# so-called **entities**, which are concatenated using underscores. You can
# access it directly:

bids_path.basename

# %%
# The two entities you can see here are the ``subject`` entity (``sub``) and
# the ``session`` entity (``ses``). Each entity name also has a value; for
# ``sub``, this is ``01``, and for ``ses``, it is ``eeg`` in our example.
# Entity names (or "keys") and values are separated via hyphens.
# BIDS knows a much larger number of entities, and MNE-BIDS allows you to make
# use of them. To get a list of all supported entities, use:

bids_path.entities

# %%
# As you can see, most entity keys are set to ``None``, which is the default
# and implies that no value has been set. Let us add a ``run`` entity, and
# remove the ``session``:

run = '01'
session = None
bids_path.update(run=run, session=session)
bids_path

# %%
# As you can see, the ``basename`` has been updated. In fact, the entire
# **path** has been updated, and the ``ses-eeg`` folder has been dropped from
# the path:

print(bids_path.fpath)

# %%
# Oups! The cell above produced a ``RuntimeWarning`` that our data file could
# not be found. That's because we changed the ``run`` and ``session`` entities
# above, and the ``tiny_bids`` dataset does not contain corresponding data.
#
# That shows us that ``BIDSPath`` is doing a lot of guess-work and checking
# in the background, but note that this may change in the future.
#
# For now, let's revert to the last working iteration of our ``bids_path``
# instance.

bids_path.update(run=None, session='eeg')
print(bids_path.fpath)

# %%
# Awesome! We're almost done! Two important things are still missing, though:
# the so-called **suffix** and the filename **extension**. Sometimes these
# terms are used interchangeably, but in BIDS, they have a very specific
# and different meaning!
#
# The **suffix** is the last part of a BIDS filename before the extension. It
# is the same as the datatype for MEG, EEG, and iEEG recordings (i.e.
# ``meg``, ``eeg``, and ``ieeg``, respectively) and ``T1w`` for T1-weighted
# MRI scans. But the suffix is also used to create the names of sidecar files
# like ``*_events.tsv``.
#
# Which brings us directly to the **extension**: the very last part of a
# filename. In MNE-BIDS, the extension contains a leading period, e.g.
# ``.tsv``.
# Let's put our new knowledge to use!

bids_path.update(suffix='eeg', extension='.vhdr')
print(bids_path.fpath)
bids_path

# %%
# By default, most MNE-BIDS functions will try to infer to correct
# suffix and extension for your data, and you don't need to specify them
# manually.



"""Code used to generate the tiny_bids dataset."""

# %%
import json
from pathlib import Path

import mne
import numpy as np

import mne_bids
from mne_bids import BIDSPath, write_raw_bids

data_path = mne.datasets.testing.data_path(download=False)
assert mne.datasets.has_dataset('testing'), 'Download testing data'
vhdr_path = data_path / "montage" / "bv_dig_test.vhdr"
captrak_path = data_path / "montage" / "captrak_coords.bvct"

mne_bids_root = Path(mne_bids.__file__).parents[1]
tiny_bids_root = mne_bids_root / "mne_bids" / "tests" / "data" / "tiny_bids"
tiny_bids_root.mkdir(exist_ok=True)

bids_path = BIDSPath(
    subject="01", task="rest", session="eeg", suffix="eeg", extension=".vhdr",
    datatype="eeg", root=tiny_bids_root
)

# %%
raw = mne.io.read_raw_brainvision(vhdr_path, preload=True)
montage = mne.channels.read_dig_captrak(captrak_path)

raw.set_channel_types(dict(ECG="ecg", HEOG="eog", VEOG="eog"))
raw.set_montage(montage)
raw.info["line_freq"] = 50

raw.info["subject_info"] = {
    "id": 1,
    "last_name": "Musterperson",
    "first_name": "Maxi",
    "middle_name": "Luka",
    "birthday": (1970, 10, 20),
    "sex": 2,
    "hand": 3,
}

# %%
# Add GSR and temperature channels
if 'GSR' not in raw.ch_names and 'Temperature' not in raw.ch_names:
    gsr_data = np.array([2.1e-6] * len(raw.times))
    temperature_data = np.array([36.5] * len(raw.times))

    gsr_and_temp_data = np.concatenate([
        np.atleast_2d(gsr_data),
        np.atleast_2d(temperature_data),
    ])
    gsr_and_temp_info = mne.create_info(
        ch_names=["GSR", "Temperature"],
        sfreq=raw.info["sfreq"],
        ch_types=["gsr", "temperature"],
    )
    gsr_and_temp_info["line_freq"] = raw.info["line_freq"]
    gsr_and_temp_info["subject_info"] = raw.info["subject_info"]
    with gsr_and_temp_info._unlock():
        gsr_and_temp_info["lowpass"] = raw.info["lowpass"]
        gsr_and_temp_info["highpass"] = raw.info["highpass"]
    gsr_and_temp_raw = mne.io.RawArray(
        data=gsr_and_temp_data,
        info=gsr_and_temp_info,
        first_samp=raw.first_samp,
    )
    raw.add_channels([gsr_and_temp_raw])
    del gsr_and_temp_raw, gsr_and_temp_data, gsr_and_temp_info

# %%
raw.set_annotations(None)
events = np.array([
    [0, 0, 1],
    [1000, 0, 2]
])
event_id = {
    "start_experiment": 1,
    "show_stimulus": 2
}

# %%
write_raw_bids(
    raw, bids_path, events=events, event_id=event_id, overwrite=True,
    allow_preload=True, format="BrainVision",
)
mne_bids.mark_channels(
    bids_path=bids_path,
    ch_names=['C3', 'C4', 'PO10', 'GSR', 'Temperature'],
    status=['good', 'good', 'bad', 'good', 'good'],
    descriptions=['resected', 'resected', 'continuously flat',
                  'left index finger', 'left ear']
)

# %%
dataset_description_json_path = tiny_bids_root / "dataset_description.json"
ds_json = json.loads(
    dataset_description_json_path.read_text(encoding="utf-8")
)

ds_json["Name"] = "tiny_bids"
ds_json["Authors"] = ["MNE-BIDS Developers", "And Friends"]

with open(dataset_description_json_path, "w", encoding="utf-8") as fout:
    json.dump(ds_json, fout, indent=4)
    fout.write("\n")



"""
.. currentmodule:: mne_bids

.. _ex-convert-mne-sample:

==========================================
02. Convert MNE sample data to BIDS format
==========================================

In this example we will use MNE-BIDS to organize the MNE sample data according
to the BIDS standard.
In a second step we will read the organized dataset using MNE-BIDS.

.. _BIDS dataset_description.json definition: https://bids-specification.readthedocs.io/en/stable/03-modality-agnostic-files.html#dataset-description
.. _ds000248 dataset_description.json: https://github.com/sappelhoff/bids-examples/blob/master/ds000248/dataset_description.json
"""  # noqa: D400 D205 E501

# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#          Teon Brooks <teon.brooks@gmail.com>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#          Richard Höchenberger <richard.hoechenberger@gmail.com>
#
# License: BSD-3-Clause

# %%
# First we import some basic Python libraries, followed by MNE-Python and its
# sample data, and then finally the MNE-BIDS functions we need for this example

import json
import os.path as op
from pprint import pprint
import shutil

import mne
from mne.datasets import sample

from mne_bids import (write_raw_bids, read_raw_bids, write_meg_calibration,
                      write_meg_crosstalk, BIDSPath, print_dir_tree,
                      make_dataset_description)
from mne_bids.stats import count_events

# %%
# Now we can read the MNE sample data. We define an `event_id` based on our
# knowledge of the data, to give meaning to events in the data.
#
# With `raw_fname` and `events`, we determine where to get the sample data
# from. `output_path` determines where we will write the BIDS conversion to.

data_path = sample.data_path()
event_id = {'Auditory/Left': 1, 'Auditory/Right': 2, 'Visual/Left': 3,
            'Visual/Right': 4, 'Smiley': 5, 'Button': 32}

raw_fname = op.join(data_path, 'MEG', 'sample', 'sample_audvis_raw.fif')
er_fname = op.join(data_path, 'MEG', 'sample', 'ernoise_raw.fif')  # empty room
events_fname = op.join(data_path, 'MEG', 'sample', 'sample_audvis_raw-eve.fif')
output_path = op.join(data_path, '..', 'MNE-sample-data-bids')

# %%
# To ensure the output path doesn't contain any leftover files from previous
# tests and example runs, we simply delete it.
#
# .. warning:: Do not delete directories that may contain important data!
#

if op.exists(output_path):
    shutil.rmtree(output_path)

# %%
#
# .. note::
#
#   ``mne-bids`` will try to infer as much information from the data as
#   possible to then save this data in BIDS-specific "sidecar" files. For
#   example the manufacturer information, which is inferred from the data file
#   extension. However, sometimes inferring is ambiguous (e.g., if your file
#   format is non-standard for the manufacturer). In these cases, MNE-BIDS does
#   *not* guess and you will have to update your BIDS fields manually.
#
# Based on our path definitions above, we read the raw data file, define
# a new BIDS name for it, and then run the automatic BIDS conversion for both
# the experimental data and its associated empty-room recording.

raw = mne.io.read_raw(raw_fname)
raw_er = mne.io.read_raw(er_fname)

# specify power line frequency as required by BIDS
raw.info['line_freq'] = 60
raw_er.info['line_freq'] = 60

task = 'audiovisual'
bids_path = BIDSPath(
    subject='01',
    session='01',
    task=task,
    run='1',
    root=output_path
)
write_raw_bids(
    raw=raw,
    bids_path=bids_path,
    events=events_fname,
    event_id=event_id,
    empty_room=raw_er,
    overwrite=True
)

# %%
# Let's pause and check that the information that we've written out to the
# sidecar files that describe our data is correct.

# Get the sidecar ``.json`` file
sidecar_json_bids_path = bids_path.copy().update(extension='.json')
sidecar_json_content = sidecar_json_bids_path.fpath.read_text(
    encoding='utf-8-sig'
)
print(sidecar_json_content)

# %%
# The sample MEG dataset comes with fine-calibration and crosstalk files that
# are required when processing Elekta/Neuromag/MEGIN data using MaxFilter®.
# Let's store these data in appropriate places, too.

cal_fname = op.join(data_path, 'SSS', 'sss_cal_mgh.dat')
ct_fname = op.join(data_path, 'SSS', 'ct_sparse_mgh.fif')

write_meg_calibration(cal_fname, bids_path)
write_meg_crosstalk(ct_fname, bids_path)

# %%
# Now let's see the structure of the BIDS folder we created.

print_dir_tree(output_path)

# %%
# Now let's get an overview of the events on the whole dataset

counts = count_events(output_path)
counts

# %%
# A big advantage of having data organized according to BIDS is that software
# packages can automate your workflow. For example, reading the data back
# into MNE-Python can easily be done using :func:`read_raw_bids`.

raw = read_raw_bids(bids_path=bids_path)

# %%
# The resulting data is already in a convenient form to create epochs and
# evoked data.

events, event_id = mne.events_from_annotations(raw)
epochs = mne.Epochs(raw, events, event_id)
epochs['Auditory'].average().plot()

# %%
# We can easily get the :class:`mne_bids.BIDSPath` of the empty-room recording
# that was associated with the experimental data while writing. The empty-room
# data can then be loaded with :func:`read_raw_bids`.

er_bids_path = bids_path.find_empty_room(use_sidecar_only=True)
er_data = read_raw_bids(er_bids_path)
er_data

# %%
# It is trivial to retrieve the path of the fine-calibration and crosstalk
# files, too.

print(bids_path.meg_calibration_fpath)
print(bids_path.meg_crosstalk_fpath)

# %%
# The README created by :func:`write_raw_bids` also takes care of the citation
# for mne-bids. If you are preparing a manuscript, please make sure to also
# cite MNE-BIDS there.
readme = op.join(output_path, 'README')
with open(readme, 'r', encoding='utf-8-sig') as fid:
    text = fid.read()
print(text)

# %%
# It is also generally a good idea to add a description of your dataset,
# see the `BIDS dataset_description.json definition`_ for more information.

how_to_acknowledge = """\
If you reference this dataset in a publication, please acknowledge its \
authors and cite MNE papers: A. Gramfort, M. Luessi, E. Larson, D. Engemann, \
D. Strohmeier, C. Brodbeck, L. Parkkonen, M. Hämäläinen, \
MNE software for processing MEG and EEG data, NeuroImage, Volume 86, \
1 February 2014, Pages 446-460, ISSN 1053-8119 \
and \
A. Gramfort, M. Luessi, E. Larson, D. Engemann, D. Strohmeier, C. Brodbeck, \
R. Goj, M. Jas, T. Brooks, L. Parkkonen, M. Hämäläinen, MEG and EEG data \
analysis with MNE-Python, Frontiers in Neuroscience, Volume 7, 2013, \
ISSN 1662-453X"""

make_dataset_description(
    path=bids_path.root,
    name=task,
    authors=["Alexandre Gramfort", "Matti Hämäläinen"],
    how_to_acknowledge=how_to_acknowledge,
    acknowledgements="""\
Alexandre Gramfort, Mainak Jas, and Stefan Appelhoff prepared and updated the \
data in BIDS format.""",
    data_license='CC0',
    ethics_approvals=['Human Subjects Division at the University of Washington'],  # noqa: E501
    funding=[
        "NIH 5R01EB009048",
        "NIH 1R01EB009048",
        "NIH R01EB006385",
        "NIH 1R01HD40712",
        "NIH 1R01NS44319",
        "NIH 2R01NS37462",
        "NIH P41EB015896",
        "ANR-11-IDEX-0003-02",
        "ERC-StG-263584",
        "ERC-StG-676943",
        "ANR-14-NEUC-0002-01"
    ],
    references_and_links=[
        "https://doi.org/10.1016/j.neuroimage.2014.02.017",
        "https://doi.org/10.3389/fnins.2013.00267",
        "https://mne.tools/stable/overview/datasets_index.html#sample"
    ],
    doi="doi:10.18112/openneuro.ds000248.v1.2.4",
    overwrite=True
)
desc_json_path = bids_path.root / 'dataset_description.json'
with open(desc_json_path, 'r', encoding='utf-8-sig') as fid:
    pprint(json.loads(fid.read()))

# %%
# This should be very similar to the `ds000248 dataset_description.json`_!



"""Count events in BIDS dataset.

example usage:  $ mne_bids count_events --bids_root bids_root_path

"""
# Authors: Alex Gramfort <alexandre.gramfort@inria.fr>
#
# License: BSD-3-Clause
from pathlib import Path

import mne_bids
from mne_bids.stats import count_events


def run():
    """Run the raw_to_bids command."""
    import pandas as pd
    from mne.commands.utils import get_optparser

    parser = get_optparser(__file__, usage="usage: %prog options args",
                           prog_prefix='mne_bids',
                           version=mne_bids.__version__)

    parser.add_option('--bids_root', dest='bids_root',
                      help='The path of the BIDS compatible folder.')

    parser.add_option('--datatype', dest='datatype', default='auto',
                      help='The datatype to consider.')

    parser.add_option('--describe', dest='describe', action="store_true",
                      help=('If set print the descriptive statistics '
                            '(min, max, etc.).'))

    parser.add_option('--output', dest='output', default=None,
                      help='Path to the CSV file where to store the results.')

    parser.add_option('--overwrite', dest='overwrite', action='store_true',
                      help='If set, overwrite an existing output file.')

    parser.add_option('--silent', dest='silent', action='store_true',
                      help='Whether to print the event counts on the screen.')

    opt, args = parser.parse_args()

    if len(args) > 0:
        parser.print_help()
        parser.error('Do not specify arguments without flags. Found: "{}".\n'
                     .format(args))

    if not all([opt.bids_root]):
        parser.print_help()
        parser.error('Arguments missing. You need to specify the '
                     '--bids_root parameter.')

    if opt.output and Path(opt.output).exists() and not opt.overwrite:
        parser.error('Output file exists. To overwrite, pass --overwrite')

    counts = count_events(opt.bids_root, datatype=opt.datatype)

    if opt.describe:
        counts = counts.describe()

    if not opt.silent:
        with pd.option_context(
            'display.max_rows', 1000,
            'display.max_columns', 80,
            'display.width', 1000
        ):
            print(counts)

    if opt.output:
        counts.to_csv(opt.output)
        print(f'\nOutput stored in {opt.output}')


if __name__ == '__main__':
    run()



"""
.. _ieeg-example:

.. currentmodule:: mne_bids

====================================
08. Convert iEEG data to BIDS format
====================================

In this example, we use MNE-BIDS to create a BIDS-compatible directory of iEEG
data. Specifically, we will follow these steps:

1. Download some iEEG data.

2. Load the data, extract information, and save in a new BIDS directory.

3. Check the result and compare it with the standard.

4. Cite MNE-BIDS.

5. Repeat the process for the ``fsaverage`` template coordinate space.

The iEEG data will be written by :func:`write_raw_bids` with
the addition of extra metadata elements in the following files:

- the sidecar file ``ieeg.json``
- ``electrodes.tsv``
- ``coordsystem.json``
- ``events.tsv``
- ``channels.tsv``

Compared to EEG data, the main differences are within the
``coordsystem.json`` and ``electrodes.tsv`` files.
For more information on these files,
refer to the `iEEG part of the BIDS specification`_.

.. _iEEG part of the BIDS specification: https://bids-specification.readthedocs.io/en/latest/04-modality-specific-files/04-intracranial-electroencephalography.html
.. _appendix VIII: https://bids-specification.readthedocs.io/en/stable/99-appendices/08-coordinate-systems.html
.. _background on FreeSurfer: https://mne.tools/dev/auto_tutorials/source-modeling/plot_background_freesurfer_mne
.. _MNE-Python coordinate frames: https://mne.tools/dev/auto_tutorials/source-modeling/plot_source_alignment.html

"""  # noqa: E501

# Authors: Adam Li <adam2392@gmail.com>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#          Alex Rockhill <aprockhill@mailbox.org>
#
# License: BSD-3-Clause

# %%

import os.path as op
import numpy as np
import shutil

import nibabel as nib
from nilearn.plotting import plot_anat

import mne
from mne_bids import (BIDSPath, write_raw_bids, write_anat,
                      get_anat_landmarks, read_raw_bids,
                      search_folder_for_text, print_dir_tree,
                      template_to_head, convert_montage_to_ras,
                      convert_montage_to_mri)

# %%
# Step 1: Download the data
# -------------------------
#
# First, we need some data to work with. We will use the
# data downloaded via MNE-Python's ``datasets`` API:
# :func:`mne.datasets.misc.data_path`
misc_path = mne.datasets.misc.data_path()

# The electrode coords data are in the tsv file format
# which is easily read in using numpy
raw = mne.io.read_raw_fif(op.join(
    misc_path, 'seeg', 'sample_seeg_ieeg.fif'))
raw.info['line_freq'] = 60  # specify power line frequency as required by BIDS
subjects_dir = op.join(misc_path, 'seeg')  # Freesurfer recon-all directory

# %%
# When the locations of the channels in this dataset were found
# in `Locating Intracranial Electrode Contacts
# <https://mne.tools/dev/auto_tutorials/clinical/10_ieeg_localize.html>`_,
# the T1 was aligned to ACPC. So, this montage is in an
# `ACPC-aligned coordinate system
# <https://surfer.nmr.mgh.harvard.edu/fswiki/CoordinateSystems>`_.
# We can either save the channel positions in the subject's anatomical
# space (from their T1 image) or we can transform to a template space
# such as ``fsaverage``. To save them in the individual space, it is
# required that the T1 have been aligned to ACPC and then the channel positions
# be in terms of that coordinate system. Automated alignment to ACPC has not
# been implemented in MNE yet, so if the channel positions are not in
# an ACPC-aligned coordinate system, using a template (like ``fsaverage``)
# is the best option.

# estimate the transformation from "head" to "mri" space
trans = mne.coreg.estimate_head_mri_t('sample_seeg', subjects_dir)

# %%
# Now let's convert the montage to "ras"
montage = raw.get_montage()
montage.apply_trans(trans)  # head->mri
convert_montage_to_ras(montage, 'sample_seeg', subjects_dir)  # mri->ras

# %%
# BIDS vs MNE-Python Coordinate Systems
# -------------------------------------
#
# BIDS has many acceptable coordinate systems for iEEG, which can be viewed in
# `appendix VIII`_ of the BIDS specification.
# However, MNE-BIDS depends on MNE-Python and MNE-Python does not support all
# these coordinate systems (yet).
#
# MNE-Python has a few tutorials on this topic:
#
# - `background on FreeSurfer`_
# - `MNE-Python coordinate frames`_
#
# MNE-Python supports using ``mni_tal`` and ``mri`` coordinate frames,
# corresponding to the ``fsaverage`` and ``ACPC`` (for an ACPC-aligned T1) BIDS
# coordinate systems respectively. All other coordinate coordinate frames in
# MNE-Python, if written with :func:`mne_bids.write_raw_bids`, must have
# an :attr:`mne_bids.BIDSPath.space` specified, and will be read in with
# the montage channel locations set to the coordinate frame 'unknown'.
#
# Step 2: Formatting as BIDS
# --------------------------
#
# Now, let us format the `Raw` object into BIDS.
#
# With this step, we have everything to start a new BIDS directory using
# our data. To do that, we can use :func:`write_raw_bids`
# Generally, :func:`write_raw_bids` tries to extract as much
# meta data as possible from the raw data and then formats it in a BIDS
# compatible way. :func:`write_raw_bids` takes a bunch of inputs, most of
# which are however optional. The required inputs are:
#
# - :code:`raw`
# - :code:`bids_basename`
# - :code:`bids_root`
#
# ... as you can see in the docstring:
print(write_raw_bids.__doc__)

# %%
# Let us initialize some of the necessary data for the subject.

# There is a subject, and specific task for the dataset.
subject_id = '1'
task = 'motor'

# get MNE-Python directory w/ example data
mne_data_dir = mne.get_config('MNE_DATASETS_MISC_PATH')

# There is the root directory for where we will write our data.
bids_root = op.join(mne_data_dir, 'ieeg_bids')

# %%
# To ensure the output path doesn't contain any leftover files from previous
# tests and example runs, we simply delete it.
#
# .. warning:: Do not delete directories that may contain important data!
#

if op.exists(bids_root):
    shutil.rmtree(bids_root)

# %%
# Now we just need make a :class:`mne_bids.BIDSPath` to save the data.
#
# .. warning:: By passing ``acpc_aligned=True``, we are affirming that
#              the T1 in this dataset is aligned to ACPC. This is very
#              difficult to check with a computer which is why this
#              step is required.

# Now convert our data to be in a new BIDS dataset.
bids_path = BIDSPath(subject=subject_id, task=task, root=bids_root)

# plot T1 to show that it is ACPC-aligned
# note that the origin is centered on the anterior commissure (AC)
# with the y-axis passing through the posterior commissure (PC)
T1_fname = op.join(subjects_dir, 'sample_seeg', 'mri', 'T1.mgz')
fig = plot_anat(T1_fname, cut_coords=(0, 0, 0))
fig.axes['x'].ax.annotate('AC', (2., -2.), (30., -40.), color='w',
                          arrowprops=dict(facecolor='w', alpha=0.5))
fig.axes['x'].ax.annotate('PC', (-31., -2.), (-80., -40.), color='w',
                          arrowprops=dict(facecolor='w', alpha=0.5))

# write ACPC-aligned T1
landmarks = get_anat_landmarks(T1_fname, raw.info, trans,
                               'sample_seeg', subjects_dir)
T1_bids_path = write_anat(T1_fname, bids_path, deface=True,
                          landmarks=landmarks)

# write `raw` to BIDS and anonymize it (converts to BrainVision format)
#
# we need to pass the `montage` argument for coordinate frames other than
# "head" which is what MNE uses internally in the `raw` object
#
# `acpc_aligned=True` affirms that our MRI is aligned to ACPC
# if this is not true, convert to `fsaverage` (see below)!
write_raw_bids(raw, bids_path, anonymize=dict(daysback=40000),
               montage=montage, acpc_aligned=True, overwrite=True)

# check our output
print_dir_tree(bids_root)

# %%
# MNE-BIDS has created a suitable directory structure for us, and among other
# meta data files, it started an ``events.tsv`` and ``channels.tsv`` file,
# and created an initial ``dataset_description.json`` file on top!
#
# Now it's time to manually check the BIDS directory and the meta files to add
# all the information that MNE-BIDS could not infer. For instance, you must
# describe ``iEEGReference`` and ``iEEGGround`` yourself.
# It's easy to find these by searching for ``"n/a"`` in the sidecar files.

search_folder_for_text('n/a', bids_root)

# %%
# Remember that there is a convenient JavaScript tool to validate all your BIDS
# directories called the "BIDS-validator", available as a web version and a
# command line tool:
#
# Web version: https://bids-standard.github.io/bids-validator/
#
# Command line tool: https://www.npmjs.com/package/bids-validator

# %%
# Step 3: Load channels from BIDS-formatted dataset and compare
# -------------------------------------------------------------
#
# Now we have written our BIDS directory. We can use
# :func:`read_raw_bids` to read in the data.

# read in the BIDS dataset to plot the coordinates
raw2 = read_raw_bids(bids_path=bids_path)

# %%
# Now we have to go back to "head" coordinates.
#
# .. note:: If you were downloading this from ``OpenNeuro``, you would
#           have to run the Freesurfer ``recon-all`` to get the transforms.

montage2 = raw2.get_montage()

# we need to go from scanner RAS back to surface RAS (requires recon-all)
convert_montage_to_mri(montage2, 'sample_seeg', subjects_dir=subjects_dir)

# this uses Freesurfer recon-all subject directory
montage2.add_estimated_fiducials('sample_seeg', subjects_dir=subjects_dir)

# get head->mri trans, invert from mri->head
trans2 = mne.transforms.invert_transform(
    mne.channels.compute_native_head_t(montage2))

# now the montage is properly in "head" and ready for analysis in MNE
raw2.set_montage(montage2)

# get the monage, apply the trans and make sure it's the same
# note: the head coordinates may differ because they are defined by
# the fiducials which are estimated; as long as the head->mri trans
# is computed with the same fiducials, the coordinates will be the same
# in ACPC space which is what matters
montage = raw.get_montage()  # the original montage in 'head' coordinates
montage.apply_trans(trans)
montage2 = raw2.get_montage()  # the recovered montage in 'head' coordinates
montage2.apply_trans(trans2)

# compare with standard
print('Recovered coordinate: {recovered}\n'
      'Saved coordinate:     {saved}'.format(
          recovered=montage2.get_positions()['ch_pos']['LENT 1'],
          saved=montage.get_positions()['ch_pos']['LENT 1']))

# %%
# Step 4: Cite mne-bids
# ---------------------
# We can see that the appropriate citations are already written in the README.
# If you are preparing a manuscript, please make sure to also cite MNE-BIDS
# there.
readme = op.join(bids_root, 'README')
with open(readme, 'r', encoding='utf-8-sig') as fid:
    text = fid.read()
print(text)

# %%
# Step 5: Store coordinates in a template space
# ---------------------------------------------
# Alternatively, if your T1 is not aligned to ACPC-space or you prefer to
# store the coordinates in a template space (e.g. ``fsaverage``) for another
# reason, you can also do that.
#
# Here we'll use the MNI Talairach transform to get to ``fsaverage`` space
# from "mri" aka surface RAS space.
# ``fsaverage`` is very useful for group analysis as shown in
# :ref:`tut-working-with-seeg`. Note, this is only a linear transform and so
# one loses quite a bit of accuracy relative to the needs of intracranial
# researchers so it is quite suboptimal. A better option is to use a
# symmetric diffeomorphic transform to create a one-to-one mapping of brain
# voxels from the individual's brain to the template as shown in
# :ref:`tut-ieeg-localize`. Even so, it's better to provide the coordinates
# in the individual's brain space, as was done above, so that the researcher
# who uses the coordinates has the ability to tranform them to a template
# of their choice.
#
# .. note::
#
#     For ``fsaverage``, the template coordinate system was defined
#     so that ``scanner RAS`` is equivalent to ``surface RAS``.
#     BIDS requires that template data be in ``scanner RAS`` so for
#     coordinate frames where this is not the case, the coordinates
#     must be converted (see below).

# ensure the output path doesn't contain any leftover files from previous
# tests and example runs
if op.exists(bids_root):
    shutil.rmtree(bids_root)

# load our raw data again
raw = mne.io.read_raw_fif(op.join(
    misc_path, 'seeg', 'sample_seeg_ieeg.fif'))
raw.info['line_freq'] = 60  # specify power line frequency as required by BIDS

# get Talairach transform
mri_mni_t = mne.read_talxfm('sample_seeg', subjects_dir)

# %%
# Now let's convert the montage to MNI Talairach ("mni_tal").
montage = raw.get_montage()
montage.apply_trans(trans)  # head->mri
montage.apply_trans(mri_mni_t)

# write to BIDS, this time with a template coordinate system
write_raw_bids(raw, bids_path, anonymize=dict(daysback=40000),
               montage=montage, overwrite=True)

# read in the BIDS dataset
raw2 = read_raw_bids(bids_path=bids_path)

# %%
# MNE-Python uses ``head`` coordinates with a ``head -> mri`` ``trans`` so we
# need to make sure to get our data in this form. As shown below, the montage
# is in the ``mni_tal`` coordinate frame but doesn't have fiducials. The
# ``head`` coordinate frame is defined based on the fiducial points so we need
# to add these. Fortunately, there is a convenient function
# (:func:`mne_bids.template_to_head`) that loads stored fiducials and takes
# care of the transformations. Once this function is applied, you can use
# the ``raw`` object and the ``trans`` as in any MNE example
# (e.g. :ref:`tut-working-with-seeg`).

# use `coord_frame='mri'` to indicate that the montage is in surface RAS
# and `unit='m'` to indicate that the units are in meters
trans2 = template_to_head(
    raw2.info, space='fsaverage', coord_frame='mri', unit='m')[1]
# this a bit confusing since we transformed from mri->mni and now we're
# saying we're back in 'mri' but that is because we were in the surface RAS
# coordinate frame of `sample_seeg` and transformed to 'mni_tal', which is the
# surface RAS coordinate frame for `fsaverage`: since MNE denotes surface RAS
# as 'mri', both coordinate frames are 'mri', it's just that 'mni_tal' is 'mri'
# when the subject is 'fsaverage'

# %%
# Let's check that we can recover the original coordinates from the BIDS
# dataset now that we are working in the ``head`` coordinate frame with a
# ``head -> mri`` ``trans`` which is the setup MNE-Python is designed around.

# check that we can recover the coordinates
print('Recovered coordinate head: {recovered}\n'
      'Original coordinate head:  {original}'.format(
          recovered=raw2.info['chs'][0]['loc'][:3],
          original=raw.info['chs'][0]['loc'][:3]))

# check difference in trans
print('Recovered trans:\n{recovered}\n'
      'Original trans:\n{original}'.format(
          recovered=trans2['trans'].round(3),
          # combine head->mri with mri->mni to get head->mni
          # and then invert to get mni->head
          original=np.linalg.inv(np.dot(trans['trans'], mri_mni_t['trans'])
                                 ).round(3)))

# ensure that the data in MNI coordinates is exactly the same
# (within computer precision)
montage2 = raw2.get_montage()  # get montage after transformed back to head
montage2.apply_trans(trans2)
print('Recovered coordinate: {recovered}\n'
      'Original coordinate:  {original}'.format(
          recovered=montage2.get_positions()['ch_pos']['LENT 1'],
          original=montage.get_positions()['ch_pos']['LENT 1']))

# %%
# As you can see the coordinates stored in the ``raw`` object are slightly off.
# This is because the ``head`` coordinate frame is defined by the fiducials
# (nasion, left and right pre-auricular points), and, in the first case,
# the fiducials were found on the individual anatomy and then transformed
# to MNI space, whereas, in the second case, they were found directly on
# the template brain (this was done once for the template so that we could
# just load it from a file). This difference means that there are slightly
# different head->mri transforms. Once these transforms are applied, however,
# the positions are the same in MNI coordinates which is what is important.
#
# As a final step, let's go over how to assign coordinate systems that are
# not recognized by MNE-Python. Many template coordinate systems are allowed by
# BIDS but are not used in MNE-Python. For these templates, the fiducials have
# been found and the transformations have been pre-computed so that we can
# get our coordinates in the ``head`` coordinate frame that MNE-Python uses.
#
# .. note::
#     As of this writing, BIDS accepts channel coordinates in reference to the
#     the following template spaces: ``ICBM452AirSpace``,
#     ``ICBM452Warp5Space``, ``IXI549Space``, ``fsaverage``, ``fsaverageSym``,
#     ``fsLR``, ``MNIColin27``, ``MNI152Lin``,
#     ``MNI152NLin2009[a-c][Sym|Asym]``, ``MNI152NLin6Sym``,
#     ``MNI152NLin6ASym``, ``MNI305``, ``NIHPD``, ``OASIS30AntsOASISAnts``,
#     ``OASIS30Atropos``, ``Talairach`` and ``UNCInfant``. As discussed above,
#     it is recommended to share the coordinates in the individual subject's
#     anatomical reference frame so that researchers who use the data can
#     transform the coordinates to any of these templates that they choose.
#
# BIDS requires that the template be stored in ``scanner RAS`` coordinates
# so first we'll convert our original data to ``scanner RAS`` and then
# convert it back. Just in case the template electrode coordinates are
# provided in voxels or the unit is not specified, these options are able
# to be overridden in :func:`mne_bids.template_to_head` for ease of use.
#
# .. warning::
#
#     If no coordinate frame is passed to :func:`mne_bids.template_to_head`
#     it will infer ``voxels`` if the coordinates are only positive and
#     ``scanner RAS`` otherwise. Be sure not to use the wrong coordinate
#     frame! ``surface RAS`` and ``scanner RAS`` are quite similar which
#     is especially confusing, but, fortunately, in most of the Freesurfer
#     template coordinate systems ``surface RAS`` is identical to
#     ``scanner RAS``. ``surface RAS`` is a Freesurfer coordinate frame so
#     it is most likely to be used with Freesurfer template coordinate
#     systems). This is the case for ``fsaverage``, ``MNI305`` and
#     ``fsaverageSym`` but not ``fsLR``.

# %%
# The template should be in scanner RAS:

# ensure the output path doesn't contain any leftover files from previous
# tests and example runs
if op.exists(bids_root):
    shutil.rmtree(bids_root)

# get a template mgz image to transform the montage to voxel coordinates
subjects_dir = op.join(mne.datasets.sample.data_path(), 'subjects')
template_T1 = nib.load(op.join(subjects_dir, 'fsaverage', 'mri', 'T1.mgz'))

# get voxels to surface RAS and scanner RAS transforms
vox_mri_t = template_T1.header.get_vox2ras_tkr()  # surface RAS
vox_ras_t = template_T1.header.get_vox2ras()  # scanner RAS

raw = mne.io.read_raw_fif(op.join(  # load our raw data again
    misc_path, 'seeg', 'sample_seeg_ieeg.fif'))
montage = raw.get_montage()  # get the original montage
montage.apply_trans(trans)  # head->mri
montage.apply_trans(mri_mni_t)  # mri->mni
pos = montage.get_positions()
ch_pos = np.array(list(pos['ch_pos'].values()))  # get an array of positions
# mri -> vox and m -> mm
ch_pos = mne.transforms.apply_trans(np.linalg.inv(vox_mri_t), ch_pos * 1000)
ch_pos = mne.transforms.apply_trans(vox_ras_t, ch_pos)

montage_ras = mne.channels.make_dig_montage(
    ch_pos=dict(zip(pos['ch_pos'].keys(), ch_pos)), coord_frame='ras')

# specify our standard template coordinate system space
bids_path.update(datatype='ieeg', space='fsaverage')

# write to BIDS, this time with a template coordinate system in voxels
write_raw_bids(raw, bids_path, anonymize=dict(daysback=40000),
               montage=montage_ras, overwrite=True)

# %%
# Now, let's load our data and convert our montage to ``head``.

raw2 = read_raw_bids(bids_path=bids_path)
trans2 = template_to_head(  # unit='auto' automatically determines it's in mm
    raw2.info, space='fsaverage', coord_frame='ras', unit='auto')[1]

# %%
# Let's check to make sure again that the original coordinates from the BIDS
# dataset were recovered.

montage2 = raw2.get_montage()  # get montage after transformed back to head
montage2.apply_trans(trans2)  # apply trans to go back to 'mri'
print('Recovered coordinate: {recovered}\n'
      'Original coordinate:  {original}'.format(
          recovered=montage2.get_positions()['ch_pos']['LENT 1'],
          original=montage.get_positions()['ch_pos']['LENT 1']))

# %%
# In summary, as we saw, these standard template spaces that are allowable by
# BIDS are quite complicated. We therefore only cover these cases because
# datasets are allowed to be in these coordinate systems, and we want to be
# able to analyze them with MNE-Python. BIDS data in a template coordinate
# space doesn't allow you to convert to a template of your choosing so it is
# better to save the raw data in the individual's ACPC space. Thus, we
# recommend, if at all possible, saving BIDS iEEG data in ACPC coordinate space
# corresponding to the individual subject's brain, not in a template
# coordinate frame.



"""MNE software for easily interacting with BIDS compatible datasets."""

__version__ = '0.11.1'
from mne_bids import commands
from mne_bids.report import make_report
from mne_bids.path import (BIDSPath, get_datatypes, get_entity_vals,
                           print_dir_tree, get_entities_from_fname,
                           search_folder_for_text, get_bids_path_from_fname)
from mne_bids.read import get_head_mri_trans, read_raw_bids
from mne_bids.utils import get_anonymization_daysback
from mne_bids.write import (make_dataset_description, write_anat,
                            write_raw_bids, mark_channels,
                            write_meg_calibration, write_meg_crosstalk,
                            get_anat_landmarks, anonymize_dataset)
from mne_bids.sidecar_updates import update_sidecar_json, update_anat_landmarks
from mne_bids.inspect import inspect_dataset
from mne_bids.dig import (template_to_head, convert_montage_to_ras,
                          convert_montage_to_mri)



"""Testing utilities for the MNE BIDS converter."""
# Authors: Chris Holdgraf <choldgraf@berkeley.edu>
#          Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#
# License: BSD-3-Clause
import os.path as op
from datetime import datetime
from pathlib import Path

import pytest
from numpy.random import random, RandomState

import mne

from mne_bids import BIDSPath
from mne_bids.utils import (_check_types, _age_on_date, _handle_datatype,
                            _infer_eeg_placement_scheme, _get_ch_type_mapping,
                            _check_datatype)
from mne_bids.path import _path_to_str

base_path = op.join(op.dirname(mne.__file__), 'io')
subject_id = '01'
session_id = '01'
run = '01'
acq = None
task = 'testing'

bids_path = BIDSPath(
    subject=subject_id, session=session_id, run=run, acquisition=acq,
    task=task)


def test_get_ch_type_mapping():
    """Test getting a correct channel mapping."""
    with pytest.raises(ValueError, match='specified from "bogus" to "mne"'):
        _get_ch_type_mapping(fro='bogus', to='mne')


def test_handle_datatype():
    """Test the automatic extraction of datatype from the data."""
    # Create a dummy raw
    n_channels = 2
    sampling_rate = 100
    data = random((n_channels, sampling_rate))
    # datatype is given, check once for each datatype
    channel_types = ['grad', 'eeg', 'ecog', 'seeg', 'dbs']
    datatypes = ['meg', 'eeg', 'ieeg', 'ieeg', 'ieeg']
    for ch_type, datatype in zip(channel_types, datatypes):
        info = mne.create_info(n_channels, sampling_rate,
                               ch_types=[ch_type] * 2)
        raw = mne.io.RawArray(data, info)
        assert _handle_datatype(raw, datatype) == datatype
    # datatype is not given, will be inferred if possible
    datatype = None
    # check if datatype is correctly inferred (combined EEG and iEEG/MEG data)
    channel_types = [['grad', 'eeg'], ['eeg', 'mag'], ['eeg', 'seeg'],
                     ['ecog', 'eeg']]
    expected_modalities = ['meg', 'meg', 'ieeg', 'ieeg']
    for ch_type, expected_mod in zip(channel_types, expected_modalities):
        info = mne.create_info(n_channels, sampling_rate, ch_types=ch_type)
        raw = mne.io.RawArray(random((2, sampling_rate)), info)
        assert _handle_datatype(raw, datatype) == expected_mod
    # set type to MEG if type is EEG/iEEG but there are MEG channels as well
    channel_types = [['grad', 'eeg'], ['grad', 'seeg']]
    datatypes = ['eeg', 'ieeg']
    for ch_type, datatype in zip(channel_types, datatypes):
        info = mne.create_info(n_channels, sampling_rate, ch_types=ch_type)
        raw = mne.io.RawArray(random((2, sampling_rate)), info)
        assert _handle_datatype(raw, datatype) == 'meg'
    # if the situation is ambiguous (iEEG and MEG), raise ValueError
    datatype = None
    channel_types = [['grad', 'ecog'], ['grad', 'seeg']]
    for ch_type in channel_types:
        with pytest.raises(ValueError, match='Multiple data types'):
            info = mne.create_info(n_channels, sampling_rate, ch_types=ch_type)
            raw = mne.io.RawArray(random((2, sampling_rate)), info)
            _handle_datatype(raw, datatype)
    # if proper channel type (iEEG, EEG or MEG) is not found, raise ValueError
    ch_type = ['misc']
    with pytest.raises(ValueError, match='No MEG, EEG or iEEG channels found'):
        info = mne.create_info(n_channels, sampling_rate,
                               ch_types=ch_type * 2)
        raw = mne.io.RawArray(data, info)
        _handle_datatype(raw, datatype)


def test_check_types():
    """Test the check whether vars are str or None."""
    assert _check_types(['foo', 'bar', None]) is None
    with pytest.raises(ValueError):
        _check_types([None, 1, 3.14, 'meg', [1, 2]])


def test_path_to_str():
    """Test that _path_to_str returns a string."""
    path_str = 'foo'
    assert _path_to_str(path_str) == path_str
    assert _path_to_str(Path(path_str)) == path_str

    with pytest.raises(ValueError):
        _path_to_str(1)


def test_age_on_date():
    """Test whether the age is determined correctly."""
    bday = datetime(1994, 1, 26)
    exp1 = datetime(2018, 1, 25)
    exp2 = datetime(2018, 1, 26)
    exp3 = datetime(2018, 1, 27)
    exp4 = datetime(1990, 1, 1)
    assert _age_on_date(bday, exp1) == 23
    assert _age_on_date(bday, exp2) == 24
    assert _age_on_date(bday, exp3) == 24
    with pytest.raises(ValueError):
        _age_on_date(bday, exp4)


def test_infer_eeg_placement_scheme():
    """Test inferring a correct EEG placement scheme."""
    # no eeg channels case (e.g., MEG data)
    data_path = op.join(base_path, 'bti', 'tests', 'data')
    raw_fname = op.join(data_path, 'test_pdf_linux')
    config_fname = op.join(data_path, 'test_config_linux')
    headshape_fname = op.join(data_path, 'test_hs_linux')
    raw = mne.io.read_raw_bti(raw_fname, config_fname, headshape_fname)
    placement_scheme = _infer_eeg_placement_scheme(raw)
    assert placement_scheme == 'n/a'

    # 1020 case
    data_path = op.join(base_path, 'brainvision', 'tests', 'data')
    raw_fname = op.join(data_path, 'test.vhdr')
    raw = mne.io.read_raw_brainvision(raw_fname)
    placement_scheme = _infer_eeg_placement_scheme(raw)
    assert placement_scheme == 'based on the extended 10/20 system'

    # Unknown case, use raw from 1020 case but rename a channel
    raw.rename_channels({'P3': 'foo'})
    placement_scheme = _infer_eeg_placement_scheme(raw)
    assert placement_scheme == 'n/a'


def test_check_datatype():
    """Test checking if datatype exists in raw data."""
    sfreq, n_points = 1024., int(1e6)
    rng = RandomState(99)
    info_eeg = mne.create_info(['ch1', 'ch2', 'ch3'], sfreq, ['eeg'] * 3)
    raw_eeg = mne.io.RawArray(rng.random((3, n_points)) * 1e-6, info_eeg)
    info_meg = mne.create_info(['ch1', 'ch2', 'ch3'], sfreq, ['mag'] * 3)
    raw_meg = mne.io.RawArray(rng.random((3, n_points)) * 1e-6, info_meg)
    info_ieeg = mne.create_info(['ch1', 'ch2', 'ch3'], sfreq, ['seeg'] * 3)
    raw_ieeg = mne.io.RawArray(rng.random((3, n_points)) * 1e-6, info_ieeg)
    # check behavior for unsupported data types
    for datatype in (None, 'anat'):
        with pytest.raises(ValueError, match=f'The specified datatype '
                                             f'{datatype} is currently not'):
            _check_datatype(raw_eeg, datatype)
    # check behavior for matching data type
    for raw, datatype in [(raw_eeg, 'eeg'), (raw_meg, 'meg'),
                          (raw_ieeg, 'ieeg')]:
        _check_datatype(raw, datatype)
    # check for missing data type
    for raw, datatype in [(raw_ieeg, 'eeg'), (raw_meg, 'eeg'),
                          (raw_ieeg, 'meg'), (raw_eeg, 'meg'),
                          (raw_meg, 'ieeg'), (raw_eeg, 'ieeg')]:
        with pytest.raises(ValueError, match=f'The specified datatype '
                                             f'{datatype} was not found'):
            _check_datatype(raw, datatype)



r"""Inspect MEG and EEG raw data, and interactively mark channels as bad.

example usage:
$ mne_bids inspect --subject_id=01 --task=experiment --session=test \
--datatype=meg --suffix=meg --bids_root=bids_root

"""
# Authors: Richard Höchenberger <richard.hoechenberger@gmail.com>
#
# License: BSD-3-Clause

from mne.utils import logger

import mne_bids
from mne_bids import BIDSPath, inspect_dataset


def run():
    """Run the mark_channels command."""
    from mne.commands.utils import get_optparser

    parser = get_optparser(__file__, usage="usage: %prog options args",
                           prog_prefix='mne_bids',
                           version=mne_bids.__version__)

    parser.add_option('--bids_root', dest='bids_root',
                      help='The path of the folder containing the BIDS '
                           'dataset')
    parser.add_option('--subject_id', dest='subject',
                      help=('Subject name'))
    parser.add_option('--session_id', dest='session',
                      help='Session name')
    parser.add_option('--task', dest='task',
                      help='Task name')
    parser.add_option('--acq', dest='acquisition',
                      help='Acquisition parameter')
    parser.add_option('--run', dest='run',
                      help='Run number')
    parser.add_option('--proc', dest='processing',
                      help='Processing label.')
    parser.add_option('--rec', dest='recording',
                      help='Recording name')
    parser.add_option('--type', dest='datatype',
                      help='Recording data type, e.g. meg, ieeg or eeg')
    parser.add_option('--suffix', dest='suffix',
                      help='The filename suffix, i.e. the last part before '
                           'the extension')
    parser.add_option('--ext', dest='extension',
                      help='The filename extension, including the leading '
                           'period, e.g. .fif')
    parser.add_option('--find_flat', dest='find_flat',
                      help='Whether to auto-detect flat channels and time '
                           'segments')
    parser.add_option('--l_freq', dest='l_freq',
                      help='The high-pass filter cutoff frequency')
    parser.add_option('--h_freq', dest='h_freq',
                      help='The low-pass filter cutoff frequency')
    parser.add_option('--verbose', dest='verbose', action='store_true',
                      help='Whether do generate additional diagnostic output')

    opt, args = parser.parse_args()
    if args:
        parser.print_help()
        parser.error(f'Please do not specify arguments without flags. '
                     f'Got: {args}.\n')

    if opt.bids_root is None:
        parser.print_help()
        parser.error('You must specify bids_root')

    bids_path = BIDSPath(subject=opt.subject, session=opt.session,
                         task=opt.task, acquisition=opt.acquisition,
                         run=opt.run, processing=opt.processing,
                         recording=opt.recording, datatype=opt.datatype,
                         suffix=opt.suffix, extension=opt.extension,
                         root=opt.bids_root)

    find_flat = True if opt.find_flat is None else bool(opt.find_flat)
    l_freq = None if opt.l_freq is None else float(opt.l_freq)
    h_freq = None if opt.h_freq is None else float(opt.h_freq)

    logger.info(f'Inspecting {bids_path.basename} …')
    inspect_dataset(bids_path=bids_path, find_flat=find_flat,
                    l_freq=l_freq, h_freq=h_freq,
                    verbose=opt.verbose)


if __name__ == '__main__':
    run()



"""
=======================================================
11. Creating BIDS-compatible folder names and filenames
=======================================================

The Brain Imaging Data Structure (BIDS) has standard conventions for file
names and folder hierarchy. MNE-BIDS comes with convenience functions if you
wish to create these files/folders on your own.

.. note::

   You may automatically convert Raw objects to BIDS-compatible files with
   ``write_raw_bids``. This example is for manually creating files/folders.
"""

# Authors: Chris Holdgraf <choldgraf@berkeley.edu>
#
# License: BSD-3-Clause

# %%
# First we will import the relevant functions

from mne_bids import BIDSPath

# %%
# Creating file names for BIDS
# ----------------------------
#
# BIDS requires a specific ordering and structure for metadata fields in
# file paths, the class `BIDSPath` allows you to specify many such
# pieces of metadata, ensuring that they are in the correct order in the
# final file path. Omitted keys will not be included in the file path.

bids_path = BIDSPath(subject='test', session='two', task='mytask',
                     suffix='events', extension='.tsv')
print(bids_path)

# %%
# You may also omit the suffix, which will result in *only* a prefix for a
# file name. This could then prepended to many more files.

bids_path = BIDSPath(subject='test', task='mytask')
print(bids_path)

# %%
# Creating folders
# ----------------
#
# You can also use MNE-BIDS to create folder hierarchies.

bids_path = BIDSPath(subject='01', session='mysession',
                     datatype='meg', root='path/to/project').mkdir()
print(bids_path.directory)



"""Configure tests."""
import os
import platform

import pytest
from mne.utils import run_subprocess


# WINDOWS issues:
# the bids-validator development version does not work properly on Windows as
# of 2019-06-25 --> https://github.com/bids-standard/bids-validator/issues/790
# As a workaround, we try to get the path to the executable from an environment
# variable VALIDATOR_EXECUTABLE ... if this is not possible we assume to be
# using the stable bids-validator and make a direct call of bids-validator
# also: for windows, shell = True is needed to call npm, bids-validator etc.
# see: https://stackoverflow.com/q/28891053/5201771
@pytest.fixture(scope="session")
def _bids_validate():
    """Fixture to run BIDS validator."""
    vadlidator_args = ['--config.error=41']
    exe = os.getenv('VALIDATOR_EXECUTABLE', 'bids-validator')

    if platform.system() == 'Windows':
        shell = True
    else:
        shell = False

    bids_validator_exe = [exe, *vadlidator_args]

    def _validate(bids_root):
        cmd = [*bids_validator_exe, bids_root]
        run_subprocess(cmd, shell=shell)

    return _validate
